from __future__ import annotations

from datetime import date, datetime

import pytest
from packaging.version import parse as vparse

import ibis.expr.datatypes as dt

np = pytest.importorskip("numpy")
pd = pytest.importorskip("pandas")


@pytest.mark.parametrize(
    ("value", "expected_dtype"),
    [
        (pd.Timedelta("5 hours"), dt.Interval(unit="h")),
        (pd.Timedelta("7 minutes"), dt.Interval(unit="m")),
        (pd.Timedelta("11 milliseconds"), dt.Interval(unit="ms")),
        (pd.Timedelta("17 nanoseconds"), dt.Interval(unit="ns")),
        # numpy types
        (np.int8(5), dt.int8),
        (np.int16(-1), dt.int16),
        (np.int32(2), dt.int32),
        (np.int64(-5), dt.int64),
        (np.uint8(5), dt.uint8),
        (np.uint16(50), dt.uint16),
        (np.uint32(500), dt.uint32),
        (np.uint64(5000), dt.uint64),
        (np.float32(5.5), dt.float32),
        (np.float64(5.55), dt.float64),
        (np.bool_(True), dt.boolean),
        (np.bool_(False), dt.boolean),
        # pandas types
        (
            pd.Timestamp("2015-01-01 12:00:00", tz="US/Eastern"),
            dt.Timestamp("US/Eastern"),
        ),
    ],
)
def test_infer_dtype(value, expected_dtype):
    assert dt.infer(value) == expected_dtype


# str, pd.Timestamp, datetime, np.datetime64, numbers.Real
@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (pd.Timestamp("2019-01-01"), datetime(2019, 1, 1)),
        (pd.Timestamp("2019-01-01 00:00:00"), datetime(2019, 1, 1)),
        (pd.Timestamp("2019-01-01 01:02:03.000004"), datetime(2019, 1, 1, 1, 2, 3, 4)),
        (np.datetime64("2019-01-01"), datetime(2019, 1, 1)),
        (np.datetime64("2019-01-01 01:02:03"), datetime(2019, 1, 1, 1, 2, 3)),
    ],
)
def test_normalize_timestamp(value, expected):
    normalized = dt.normalize(dt.timestamp, value)
    assert normalized == expected


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (pd.Timestamp("2019-01-01"), date(2019, 1, 1)),
        (pd.Timestamp("2019-01-01 00:00:00"), date(2019, 1, 1)),
        (pd.Timestamp("2019-01-01 01:02:03.000004"), date(2019, 1, 1)),
        (np.datetime64("2019-01-01"), date(2019, 1, 1)),
        (np.datetime64("2019-01-01 01:02:03"), date(2019, 1, 1)),
    ],
)
def test_normalize_date(value, expected):
    normalized = dt.normalize(dt.date, value)
    assert normalized == expected


@pytest.mark.parametrize(
    ("value", "expected_dtype"),
    [
        # numpy types
        (np.int8(5), dt.int8),
        (np.int16(-1), dt.int16),
        (np.int32(2), dt.int32),
        (np.int64(-5), dt.int64),
        (np.uint8(5), dt.uint8),
        (np.uint16(50), dt.uint16),
        (np.uint32(500), dt.uint32),
        (np.uint64(5000), dt.uint64),
        (np.float32(5.5), dt.float32),
        (np.float64(5.55), dt.float64),
        (np.bool_(True), dt.boolean),
        (np.bool_(False), dt.boolean),
        # pandas types
        (
            pd.Timestamp("2015-01-01 12:00:00", tz="US/Eastern"),
            dt.Timestamp("US/Eastern"),
        ),
    ],
)
def test_infer_numpy_scalar(value, expected_dtype):
    assert dt.infer(value) == expected_dtype


@pytest.mark.parametrize(
    ("numpy_dtype", "ibis_dtype"),
    [
        (np.bool_, dt.boolean),
        (np.int8, dt.int8),
        (np.int16, dt.int16),
        (np.int32, dt.int32),
        (np.int64, dt.int64),
        (np.uint8, dt.uint8),
        (np.uint16, dt.uint16),
        (np.uint32, dt.uint32),
        (np.uint64, dt.uint64),
        (np.float16, dt.float16),
        (np.float32, dt.float32),
        (np.float64, dt.float64),
        (np.double, dt.double),
        (np.str_, dt.string),
        (np.datetime64, dt.timestamp),
    ],
)
def test_from_numpy_dtype(numpy_dtype, ibis_dtype):
    numpy_dtype = np.dtype(numpy_dtype)
    assert dt.DataType.from_numpy(numpy_dtype) == ibis_dtype
    assert dt.dtype(numpy_dtype) == ibis_dtype


def test_from_numpy_timedelta():
    if vparse(pytest.importorskip("pyarrow").__version__) < vparse("9"):
        pytest.skip("pyarrow < 9 globally mutates the timedelta64 numpy dtype")

    numpy_dtype = np.dtype(np.timedelta64)
    assert dt.DataType.from_numpy(numpy_dtype) == dt.Interval("s")
    assert dt.dtype(numpy_dtype) == dt.Interval("s")


@pytest.mark.parametrize(
    ("numpy_array", "expected_dtypes"),
    [
        # Explicitly-defined dtype
        (np.array([1, 2, 3], dtype="int8"), (dt.Array(dt.int8),)),
        (np.array([1, 2, 3], dtype="int16"), (dt.Array(dt.int16),)),
        (np.array([1, 2, 3], dtype="int32"), (dt.Array(dt.int32),)),
        (np.array([1, 2, 3], dtype="int64"), (dt.Array(dt.int64),)),
        (np.array([1, 2, 3], dtype="uint8"), (dt.Array(dt.uint8),)),
        (np.array([1, 2, 3], dtype="uint16"), (dt.Array(dt.uint16),)),
        (np.array([1, 2, 3], dtype="uint32"), (dt.Array(dt.uint32),)),
        (np.array([1, 2, 3], dtype="uint64"), (dt.Array(dt.uint64),)),
        (np.array([1.0, 2.0, 3.0], dtype="float32"), (dt.Array(dt.float32),)),
        (np.array([1.0, 2.0, 3.0], dtype="float64"), (dt.Array(dt.float64),)),
        (np.array([True, False, True], dtype="bool"), (dt.Array(dt.boolean),)),
        # Implicit dtype
        # Integer array could be inferred to int64 or int32 depending on system
        (np.array([1, 2, 3]), (dt.Array(dt.int64), dt.Array(dt.int32))),
        (np.array([1.0, 2.0, 3.0]), (dt.Array(dt.float64),)),
        (np.array([np.nan, np.nan, np.nan]), (dt.Array(dt.float64),)),
        (np.array([True, False, True]), (dt.Array(dt.boolean),)),
        (np.array(["1", "2", "3"]), (dt.Array(dt.string),)),
        (
            np.array(
                [
                    pd.Timestamp("2015-01-01 12:00:00"),
                    pd.Timestamp("2015-01-02 12:00:00"),
                    pd.Timestamp("2015-01-03 12:00:00"),
                ]
            ),
            (dt.Array(dt.Timestamp()), dt.Array(dt.Timestamp(scale=6))),
        ),
        # Implied from object dtype
        (np.array([1, 2, 3], dtype=object), (dt.Array(dt.int64),)),
        (np.array([1.0, 2.0, 3.0], dtype=object), (dt.Array(dt.float64),)),
        (np.array([True, False, True], dtype=object), (dt.Array(dt.boolean),)),
        (np.array(["1", "2", "3"], dtype=object), (dt.Array(dt.string),)),
        (
            np.array(
                [
                    pd.Timestamp("2015-01-01 12:00:00"),
                    pd.Timestamp("2015-01-02 12:00:00"),
                    pd.Timestamp("2015-01-03 12:00:00"),
                ],
                dtype=object,
            ),
            (dt.Array(dt.Timestamp()), dt.Array(dt.Timestamp(scale=6))),
        ),
    ],
)
def test_infer_numpy_array(numpy_array, expected_dtypes):
    pytest.importorskip("pyarrow")
    pandas_series = pd.Series(numpy_array)
    assert dt.infer(numpy_array) in expected_dtypes
    assert dt.infer(pandas_series) in expected_dtypes


def test_normalize_non_convertible_boolean():
    typ = dt.boolean
    value = np.array([1, 2, 3])
    with pytest.raises(TypeError, match="Unable to normalize .+ to Boolean"):
        dt.normalize(typ, value)



from __future__ import annotations

import os
from urllib.parse import quote_plus

import pandas as pd
import pandas.testing as tm
import pyarrow as pa
import pytest
from pytest import param

import ibis
import ibis.common.exceptions as exc
import ibis.expr.datatypes as dt
import ibis.expr.types as ir
from ibis import udf
from ibis.backends.clickhouse.tests.conftest import (
    CLICKHOUSE_HOST,
    CLICKHOUSE_PASS,
    CLICKHOUSE_PORT,
    CLICKHOUSE_USER,
    IBIS_TEST_CLICKHOUSE_DB,
)
from ibis.backends.tests.errors import ClickHouseDatabaseError
from ibis.util import gen_name

cc = pytest.importorskip("clickhouse_connect")


def test_connection(con):
    assert con is not None
    assert isinstance(con, ibis.backends.clickhouse.Backend)


def test_run_sql(con):
    query = "SELECT * FROM ibis_testing.functional_alltypes"
    table = con.sql(query)

    fa = con.table("functional_alltypes")
    assert isinstance(table, ir.Table)
    assert table.schema() == fa.schema()

    expr = table.limit(10)
    result = expr.execute()
    assert len(result) == 10


def test_get_schema(con):
    t = con.table("functional_alltypes")
    schema = con.get_schema("functional_alltypes")
    assert t.schema() == schema


def test_result_as_dataframe(con, alltypes):
    expr = alltypes.limit(10)

    ex_names = list(expr.schema().names)
    result = con.execute(expr)

    assert isinstance(result, pd.DataFrame)
    assert result.columns.tolist() == ex_names
    assert len(result) == 10


def test_array_default_limit(con, alltypes):
    result = con.execute(alltypes.float_col, limit=100)
    assert len(result) == 100


def test_limit_overrides_expr(con, alltypes):
    result = con.execute(alltypes.limit(10), limit=5)
    assert len(result) == 5


def test_limit_equals_none_no_limit(alltypes, monkeypatch):
    monkeypatch.setattr(ibis.options.sql, "default_limit", 10)
    result = alltypes.execute(limit=None)
    assert len(result) > 10


def test_verbose_log_queries(con, monkeypatch):
    queries = []

    monkeypatch.setattr(ibis.options, "verbose", True)
    monkeypatch.setattr(ibis.options, "verbose_log", queries.append)

    con.table("functional_alltypes")

    expected = "DESCRIBE functional_alltypes"

    # might be other queries in there, we only check that a describe table
    # query was logged
    assert expected in queries


def test_sql_query_limits_big(alltypes, monkeypatch):
    monkeypatch.setattr(ibis.options.sql, "default_limit", 100_000)

    # alltypes has 7300 rows
    assert len(alltypes.execute()) == 7300  # comply with limit arg for alltypes
    assert len(alltypes.execute(limit=10)) == 10
    # state hasn't changed
    assert len(alltypes.execute()) == 7300
    # non-alltypes ignores default_limit
    assert alltypes.count().execute() == 7300
    # non-alltypes doesn't observe limit arg
    assert alltypes.count().execute(limit=10) == 7300


def test_sql_query_limits_small(alltypes, monkeypatch):
    monkeypatch.setattr(ibis.options.sql, "default_limit", 20)

    # alltypes observes default limit setting
    assert len(alltypes.execute()) == 20
    # explicit limit= overrides default
    assert len(alltypes.execute(limit=15)) == 15
    assert len(alltypes.execute(limit=23)) == 23
    # non-alltypes ignores default_limit
    assert alltypes.count().execute() == 7300
    # non-alltypes doesn't observe limit arg
    assert alltypes.count().execute(limit=10) == 7300


def test_sql_query_limits_none(alltypes, monkeypatch):
    monkeypatch.setattr(ibis.options.sql, "default_limit", None)

    # eliminating default_limit doesn't break anything
    assert len(alltypes.execute()) == 7300
    assert len(alltypes.execute(limit=15)) == 15
    assert len(alltypes.execute(limit=10000)) == 7300
    assert alltypes.count().execute() == 7300
    assert alltypes.count().execute(limit=10) == 7300


def test_embedded_identifier_quoting(alltypes):
    t = alltypes

    expr = t.select((t.double_col * 2).name("double(fun)"))["double(fun)"].sum()
    expr.execute()


@pytest.fixture
def temporary_alltypes(con):
    table = gen_name("temporary_alltypes")
    con.raw_sql(f"CREATE TABLE {table} AS functional_alltypes")
    yield con.table(table)
    con.drop_table(table)


@pytest.mark.parametrize(
    "transform",
    [param(lambda df: df, id="pandas"), param(pa.Table.from_pandas, id="pyarrow")],
)
def test_insert(con, temporary_alltypes, df, transform):
    temporary = temporary_alltypes
    records = df[:10]

    assert temporary.count().execute() == 0

    con.insert(temporary.op().name, transform(records))

    tm.assert_frame_equal(temporary.execute(), records)


def test_insert_with_less_columns(con, temporary_alltypes, df):
    temporary = temporary_alltypes
    records = df.loc[:10, ["string_col"]].copy()
    records["date_col"] = None

    with pytest.raises(cc.driver.exceptions.DatabaseError):
        con.insert(temporary.op().name, records)


def test_insert_with_more_columns(con, temporary_alltypes, df):
    temporary = temporary_alltypes
    records = df[:10].copy()
    records["non_existing_column"] = "raise on me"

    with pytest.raises(cc.driver.exceptions.DatabaseError):
        con.insert(temporary.op().name, records)


@pytest.mark.parametrize(
    ("query", "expected_schema"),
    [
        (
            "SELECT 1 as a, 2 + dummy as b",
            ibis.schema(dict(a=dt.UInt8(nullable=False), b=dt.UInt16(nullable=False))),
        ),
        (
            "SELECT string_col, sum(double_col) as b FROM functional_alltypes GROUP BY string_col",
            ibis.schema(
                dict(
                    string_col=dt.String(nullable=True),
                    b=dt.Float64(nullable=True),
                )
            ),
        ),
    ],
)
def test_get_schema_using_query(con, query, expected_schema):
    result = con._get_schema_using_query(query)
    assert result == expected_schema


def test_list_tables_database(con):
    tables = con.list_tables()
    tables2 = con.list_tables(database=con.current_database)
    # some overlap, but not necessarily identical because
    # a database may have temporary tables added/removed between list_tables
    # calls
    assert set(tables) & set(tables2)


@pytest.fixture
def tmpcon(worker_id):
    dbname = f"clickhouse_database_{worker_id}"
    con = ibis.clickhouse.connect(
        host=os.environ.get("IBIS_TEST_CLICKHOUSE_HOST", "localhost"),
        user=os.environ.get("IBIS_TEST_CLICKHOUSE_USER", "default"),
        port=int(os.environ.get("IBIS_TEST_CLICKHOUSE_PORT", 8123)),
        password=os.environ.get("IBIS_TEST_CLICKHOUSE_PASSWORD", ""),
    )
    con.create_database(dbname, force=True)
    yield con
    con.drop_database(dbname, force=True)


def test_list_tables_empty_database(tmpcon):
    assert not tmpcon.list_tables()


@pytest.mark.parametrize("temp", [True, False], ids=["temp", "no_temp"])
def test_create_table_no_data(con, temp, temp_table):
    schema = ibis.schema(dict(a="!int", b="string"))
    t = con.create_table(temp_table, schema=schema, temp=temp, engine="Memory")
    assert t.execute().empty


@pytest.mark.parametrize(
    "data",
    [
        {"a": [1, 2, 3], "b": [None, "b", "c"]},
        pd.DataFrame({"a": [1, 2, 3], "b": [None, "b", "c"]}),
        pa.Table.from_pydict({"a": [1, 2, 3], "b": [None, "b", "c"]}),
    ],
    ids=["dict", "dataframe", "pyarrow table"],
)
@pytest.mark.parametrize(
    "engine",
    ["File(Native)", "File(Parquet)", "Memory"],
    ids=["native", "mem", "parquet"],
)
def test_create_table_data(con, data, engine, temp_table):
    schema = ibis.schema(dict(a="!int", b="string"))
    t = con.create_table(temp_table, obj=data, schema=schema, engine=engine)
    assert len(t.execute()) == 3


def test_create_table_with_properties(con, temp_table):
    data = pd.DataFrame({"a": list("abcde" * 20), "b": [1, 2, 3, 4, 5] * 20})
    n = len(data)
    t = con.create_table(
        temp_table,
        data,
        schema=ibis.schema(dict(a="string", b="!uint32")),
        order_by=["a", "b"],
        partition_by=["a"],
        sample_by=["b"],
        settings={"allow_nullable_key": "1"},
    )
    assert t.count().execute() == n


@pytest.mark.parametrize(
    "engine",
    [
        "File(Native)",
        param(
            "File(Parquet)",
            marks=pytest.mark.xfail(
                reason="Parquet file size is 0 bytes",
                raises=cc.driver.exceptions.DatabaseError,
            ),
        ),
        "Memory",
    ],
    ids=["native", "mem", "parquet"],
)
def test_truncate_table(con, engine, temp_table):
    t = con.create_table(temp_table, obj={"a": [1]}, engine=engine)
    assert len(t.execute()) == 1
    con.truncate_table(temp_table)
    assert t.execute().empty


@udf.scalar.builtin(name="arrayJaccardIndex")
def array_jaccard_index(a: dt.Array[dt.int64], b: dt.Array[dt.int64]) -> float: ...


@udf.scalar.builtin(name="arrayJaccardIndex")
def array_jaccard_index_no_input_types(a, b) -> float: ...


@udf.scalar.builtin
def arrayJaccardIndex(a: dt.Array[dt.int64], b: dt.Array[dt.int64]) -> float: ...


@pytest.mark.parametrize(
    "func",
    [
        array_jaccard_index,
        arrayJaccardIndex,
        array_jaccard_index_no_input_types,
    ],
)
def test_builtin_scalar_udf(con, func):
    expr = func([1, 2], [2, 3])
    result = con.execute(expr)
    expected = 1.0 / 3.0
    assert result == expected


@udf.agg.builtin
def entropy(a) -> float: ...


@udf.agg.builtin(name="sumKahan")
def sum_kahan(a: float) -> float: ...


@pytest.mark.parametrize("func", [entropy, sum_kahan])
def test_builtin_agg_udf(con, func):
    t = ibis.memtable(pd.DataFrame({"a": list(map(float, range(10)))}))

    expr = func(t.a)
    result = con.execute(expr)

    table_name = gen_name("agg_udf_test")
    expected = con.con.query_df(
        f"SELECT {expr.op().__func_name__}(a) FROM {table_name}",
        external_data=con._normalize_external_tables({table_name: t.op()}),
    ).squeeze()
    assert result == expected


def test_create_table_no_syntax_error(con):
    schema = ibis.schema(
        [
            ("play_id", "!string"),
            ("uid", "!int64"),
            ("feat1", "float64"),
            ("feat2", "float64"),
            ("feat3", "float64"),
            ("feat4", "float64"),
            ("feat5", "float64"),
            ("feat6", "float64"),
            ("feat7", "float64"),
            ("feat8", "float64"),
            ("feat9", "float64"),
            ("feat10", "float64"),
            ("feat11", "float64"),
            ("play_timestamp", "!int64"),
            ("tag_date", "!int32"),
        ]
    )
    t = con.create_table(gen_name("clickouse_temp_table"), schema=schema, temp=True)
    assert t.count().execute() == 0


def test_password_with_bracket():
    password = f'{os.environ.get("IBIS_TEST_CLICKHOUSE_PASSWORD", "")}[]'
    quoted_pass = quote_plus(password)
    host = os.environ.get("IBIS_TEST_CLICKHOUSE_HOST", "localhost")
    user = os.environ.get("IBIS_TEST_CLICKHOUSE_USER", "default")
    port = int(os.environ.get("IBIS_TEST_CLICKHOUSE_PORT", 8123))
    with pytest.raises(
        cc.driver.exceptions.DatabaseError, match="password is incorrect"
    ):
        ibis.clickhouse.connect(host=host, user=user, port=port, password=quoted_pass)


def test_from_url(con):
    assert ibis.connect(
        f"clickhouse://{CLICKHOUSE_USER}:{CLICKHOUSE_PASS}@{CLICKHOUSE_HOST}:{CLICKHOUSE_PORT}/{IBIS_TEST_CLICKHOUSE_DB}"
    )


def test_from_url_with_kwargs(con):
    # since explicit kwargs take precedence, this passes, because we're passing
    # `database` explicitly, even though our connection string says to use a
    # random database
    database = ibis.util.gen_name("clickhouse_database")
    assert ibis.connect(
        f"clickhouse://{CLICKHOUSE_USER}:{CLICKHOUSE_PASS}@{CLICKHOUSE_HOST}:{CLICKHOUSE_PORT}/{database}",
        database=IBIS_TEST_CLICKHOUSE_DB,
    )


def test_invalid_port():
    port = 9999
    url = f"clickhouse://{CLICKHOUSE_USER}:{CLICKHOUSE_PASS}@{CLICKHOUSE_HOST}:{port}/{IBIS_TEST_CLICKHOUSE_DB}"
    with pytest.raises(cc.driver.exceptions.DatabaseError):
        ibis.connect(url)


def test_subquery_with_join(con):
    name = gen_name("clickhouse_tmp_table")

    s = con.create_table(name, pa.Table.from_pydict({"a": [1, 2, 3]}), temp=True)

    sql = f"""
    SELECT
      "o"."a"
    FROM (
      SELECT
        "w"."a"
      FROM "{name}" AS "s"
      INNER JOIN "{name}" AS "w"
      USING ("a")
    ) AS "o"
    """
    with pytest.raises(
        ClickHouseDatabaseError, match="Identifier 'o.a' cannot be resolved"
    ):
        # https://github.com/ClickHouse/ClickHouse/issues/66133
        con.sql(sql)

    # this works because we add the additional alias in the inner query
    w = s.view()
    expr = s.join(w, "a").select(a=w.a).select(b=lambda t: t.a + 1)
    result = expr.to_pandas()
    assert set(result["b"].tolist()) == {2, 3, 4}


def test_alias_column_ref(con):
    data = {"user_id": [1, 2, 3], "account_id": [4, 5, 6]}
    t = con.create_table(gen_name("clickhouse_temp_table"), data, temp=True)
    expr = t.alias("df").sql("select *, halfMD5(account_id) as id_md5 from df")

    result = expr.execute()

    assert len(result) == 3

    assert result.columns.tolist() == ["user_id", "account_id", "id_md5"]

    assert result.user_id.notnull().all()
    assert result.account_id.notnull().all()
    assert result.id_md5.notnull().all()


@pytest.mark.parametrize("method_name", ["to_pandas", "to_pyarrow"])
def test_query_cache(con, method_name):
    t = con.table("functional_alltypes")
    expr = t.count()

    method = getattr(expr, method_name)

    expected = method()
    result = method(settings={"use_query_cache": True})

    # test a bogus setting
    with pytest.raises(ClickHouseDatabaseError):
        method(settings={"ooze_query_cash": True})

    assert result == expected


def test_invalid_catalog_argument(con):
    with pytest.raises(
        exc.UnsupportedOperationError, match="`catalog` namespaces are not supported"
    ):
        con.get_schema("t", catalog="a", database="b")



from __future__ import annotations

import pytest


@pytest.fixture(scope="session")
def customer(backend):
    return backend.h("customer")


@pytest.fixture(scope="session")
def lineitem(backend):
    return backend.h("lineitem")


@pytest.fixture(scope="session")
def nation(backend):
    return backend.h("nation")


@pytest.fixture(scope="session")
def orders(backend):
    return backend.h("orders")


@pytest.fixture(scope="session")
def part(backend):
    return backend.h("part")


@pytest.fixture(scope="session")
def partsupp(backend):
    return backend.h("partsupp")


@pytest.fixture(scope="session")
def region(backend):
    return backend.h("region")


@pytest.fixture(scope="session")
def supplier(backend):
    return backend.h("supplier")



from __future__ import annotations

try:
    from duckdb import BinderException as DuckDBBinderException
    from duckdb import ConversionException as DuckDBConversionException
    from duckdb import InvalidInputException as DuckDBInvalidInputException
    from duckdb import NotImplementedException as DuckDBNotImplementedException
    from duckdb import OutOfRangeException as DuckDBOutOfRangeException
    from duckdb import ParserException as DuckDBParserException
except ImportError:
    DuckDBConversionException = DuckDBInvalidInputException = DuckDBParserException = (
        DuckDBNotImplementedException
    ) = DuckDBBinderException = DuckDBOutOfRangeException = None

try:
    from clickhouse_connect.driver.exceptions import (
        DatabaseError as ClickHouseDatabaseError,
    )
    from clickhouse_connect.driver.exceptions import (
        InternalError as ClickHouseInternalError,
    )
    from clickhouse_connect.driver.exceptions import (
        OperationalError as ClickHouseOperationalError,
    )
except ImportError:
    ClickHouseDatabaseError = ClickHouseInternalError = ClickHouseOperationalError = (
        None
    )

try:
    from pyexasol.exceptions import ExaQueryError
except ImportError:
    ExaQueryError = None

try:
    from pyspark.sql.utils import AnalysisException as PySparkAnalysisException
    from pyspark.sql.utils import (
        IllegalArgumentException as PySparkIllegalArgumentException,
    )
    from pyspark.sql.utils import ParseException as PySparkParseException
    from pyspark.sql.utils import PythonException as PySparkPythonException
except ImportError:
    PySparkAnalysisException = PySparkIllegalArgumentException = (
        PySparkParseException
    ) = PySparkPythonException = None

try:
    # PySpark 3.5.0
    from pyspark.errors.exceptions.base import (
        AnalysisException as PySparkAnalysisException,
    )
    from pyspark.errors.exceptions.base import (
        ArithmeticException as PySparkArithmeticException,
    )
    from pyspark.errors.exceptions.base import ParseException as PySparkParseException
    from pyspark.errors.exceptions.base import PySparkValueError
    from pyspark.errors.exceptions.base import PythonException as PySparkPythonException
    from pyspark.errors.exceptions.connect import (
        SparkConnectGrpcException as PySparkConnectGrpcException,
    )
except ImportError:
    PySparkParseException = PySparkAnalysisException = PySparkArithmeticException = (
        PySparkPythonException
    ) = PySparkConnectGrpcException = PySparkValueError = None

try:
    from google.api_core.exceptions import BadRequest as GoogleBadRequest
    from google.api_core.exceptions import NotFound as GoogleNotFound
except ImportError:
    GoogleBadRequest = GoogleNotFound = None

try:
    from polars.exceptions import ColumnNotFoundError as PolarsColumnNotFoundError
    from polars.exceptions import ComputeError as PolarsComputeError
    from polars.exceptions import InvalidOperationError as PolarsInvalidOperationError
    from polars.exceptions import PanicException as PolarsPanicException
    from polars.exceptions import SchemaError as PolarsSchemaError
    from polars.exceptions import SQLInterfaceError as PolarsSQLInterfaceError
except ImportError:
    PolarsComputeError = PolarsPanicException = PolarsInvalidOperationError = (
        PolarsSchemaError
    ) = PolarsColumnNotFoundError = PolarsSQLInterfaceError = None

try:
    from pyarrow import ArrowInvalid, ArrowNotImplementedError
except ImportError:
    ArrowInvalid = ArrowNotImplementedError = None

try:
    from impala.error import HiveServer2Error as ImpalaHiveServer2Error
    from impala.error import OperationalError as ImpalaOperationalError
except ImportError:
    ImpalaHiveServer2Error = ImpalaOperationalError = None

try:
    from py4j.protocol import Py4JError, Py4JJavaError
except ImportError:
    Py4JJavaError = Py4JError = None

try:
    from deltalake import PyDeltaTableError
except ImportError:
    PyDeltaTableError = None

try:
    from snowflake.connector.errors import ProgrammingError as SnowflakeProgrammingError
except ImportError:
    SnowflakeProgrammingError = None

try:
    from trino.exceptions import TrinoUserError
except ImportError:
    TrinoUserError = None

try:
    from psycopg2.errors import ArraySubscriptError as PsycoPg2ArraySubscriptError
    from psycopg2.errors import DivisionByZero as PsycoPg2DivisionByZero
    from psycopg2.errors import IndeterminateDatatype as PsycoPg2IndeterminateDatatype
    from psycopg2.errors import InternalError_ as PsycoPg2InternalError
    from psycopg2.errors import (
        InvalidTextRepresentation as PsycoPg2InvalidTextRepresentation,
    )
    from psycopg2.errors import OperationalError as PsycoPg2OperationalError
    from psycopg2.errors import ProgrammingError as PsycoPg2ProgrammingError
    from psycopg2.errors import SyntaxError as PsycoPg2SyntaxError
    from psycopg2.errors import UndefinedObject as PsycoPg2UndefinedObject
except ImportError:
    PsycoPg2SyntaxError = PsycoPg2IndeterminateDatatype = (
        PsycoPg2InvalidTextRepresentation
    ) = PsycoPg2DivisionByZero = PsycoPg2InternalError = PsycoPg2ProgrammingError = (
        PsycoPg2OperationalError
    ) = PsycoPg2UndefinedObject = PsycoPg2ArraySubscriptError = None

try:
    from MySQLdb import NotSupportedError as MySQLNotSupportedError
    from MySQLdb import OperationalError as MySQLOperationalError
    from MySQLdb import ProgrammingError as MySQLProgrammingError
except ImportError:
    MySQLNotSupportedError = MySQLProgrammingError = MySQLOperationalError = None

try:
    from pydruid.db.exceptions import ProgrammingError as PyDruidProgrammingError
except ImportError:
    PyDruidProgrammingError = None

try:
    from oracledb.exceptions import DatabaseError as OracleDatabaseError
    from oracledb.exceptions import InterfaceError as OracleInterfaceError
except ImportError:
    OracleDatabaseError = OracleInterfaceError = None

try:
    from pyodbc import DataError as PyODBCDataError
    from pyodbc import ProgrammingError as PyODBCProgrammingError
except ImportError:
    PyODBCProgrammingError = PyODBCDataError = None

try:
    from databricks.sql.exc import (
        ServerOperationError as DatabricksServerOperationError,
    )
except ImportError:
    DatabricksServerOperationError = None



import ibis


lineitem = ibis.table(
    name="lineitem",
    schema={
        "l_orderkey": "int32",
        "l_partkey": "int32",
        "l_suppkey": "int32",
        "l_linenumber": "int32",
        "l_quantity": "decimal(15, 2)",
        "l_extendedprice": "decimal(15, 2)",
        "l_discount": "decimal(15, 2)",
        "l_tax": "decimal(15, 2)",
        "l_returnflag": "string",
        "l_linestatus": "string",
        "l_shipdate": "date",
        "l_commitdate": "date",
        "l_receiptdate": "date",
        "l_shipinstruct": "string",
        "l_shipmode": "string",
        "l_comment": "string",
    },
)
lit = ibis.literal(1)
f = lineitem.filter((lineitem.l_shipdate <= ibis.literal("1998-09-02").cast("date")))
multiply = f.l_extendedprice * ((lit - f.l_discount))
agg = f.aggregate(
    [
        f.l_quantity.sum().name("sum_qty"),
        f.l_extendedprice.sum().name("sum_base_price"),
        multiply.sum().name("sum_disc_price"),
        ((multiply) * ((lit + f.l_tax))).sum().name("sum_charge"),
        f.l_quantity.mean().name("avg_qty"),
        f.l_extendedprice.mean().name("avg_price"),
        f.l_discount.mean().name("avg_disc"),
        f.count().name("count_order"),
    ],
    by=[f.l_returnflag, f.l_linestatus],
)

result = agg.order_by(
    agg.l_returnflag.asc(nulls_first=True), agg.l_linestatus.asc(nulls_first=True)
)



import ibis


airlines = ibis.table(
    name="airlines", schema={"dest": "string", "origin": "string", "arrdelay": "int32"}
)

result = airlines.filter((((airlines.dest.cast("int64") == 0)) == True))



from __future__ import annotations

import builtins
import contextlib
import importlib
import inspect
import json
import os
import re
import string
import subprocess
import sys
from operator import itemgetter
from typing import TYPE_CHECKING

import pytest
import rich.console
import sqlglot as sg
import toolz
from packaging.version import parse as vparse
from pytest import mark, param

import ibis
import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
from ibis.backends.conftest import ALL_BACKENDS
from ibis.backends.tests.errors import (
    DatabricksServerOperationError,
    ExaQueryError,
    ImpalaHiveServer2Error,
    OracleDatabaseError,
    PsycoPg2InternalError,
    PsycoPg2UndefinedObject,
    Py4JJavaError,
    PyODBCProgrammingError,
    SnowflakeProgrammingError,
)
from ibis.conftest import IS_SPARK_REMOTE
from ibis.util import gen_name

if TYPE_CHECKING:
    from ibis.backends import BaseBackend


np = pytest.importorskip("numpy")
pd = pytest.importorskip("pandas")
pa = pytest.importorskip("pyarrow")
ds = pytest.importorskip("pyarrow.dataset")


@pytest.fixture
def new_schema():
    return ibis.schema([("a", "string"), ("b", "bool"), ("c", "int32")])


def _create_temp_table_with_schema(backend, con, temp_table_name, schema, data=None):
    if con.name == "druid":
        pytest.xfail("druid doesn't implement create_table")
    elif con.name == "flink":
        pytest.xfail(
            "flink doesn't implement create_table from schema without additional arguments"
        )
    temporary = con.create_table(temp_table_name, schema=schema)
    assert temporary.to_pandas().empty

    if data is not None and isinstance(data, pd.DataFrame):
        assert not data.empty
        tmp = con.create_table(temp_table_name, data, overwrite=True)
        result = tmp.to_pandas()
        assert len(result) == len(data.index)
        backend.assert_frame_equal(
            result.sort_values(result.columns[0]).reset_index(drop=True),
            data.sort_values(result.columns[0]).reset_index(drop=True),
        )
        return tmp

    return temporary


@pytest.mark.parametrize(
    "func", [toolz.identity, pa.Table.from_pandas], ids=["dataframe", "pyarrow_table"]
)
@pytest.mark.parametrize(
    "sch",
    [
        None,
        dict(first_name="string", salary="float64"),
        dict(first_name="string", salary="float64").items(),
        ibis.schema(dict(first_name="string", salary="float64")),
    ],
    ids=["no_schema", "dict_schema", "tuples", "schema"],
)
@pytest.mark.notimpl(["druid"])
@pytest.mark.notimpl(
    ["flink"],
    reason="Flink backend supports creating only TEMPORARY VIEW for in-memory data.",
)
def test_create_table(backend, con, temp_table, func, sch):
    df = pd.DataFrame({"first_name": ["A", "B", "C"], "salary": [100.0, 200.0, 300.0]})

    con.create_table(temp_table, func(df), schema=sch)
    result = (
        con.table(temp_table).execute().sort_values("first_name").reset_index(drop=True)
    )

    backend.assert_frame_equal(df, result)


@pytest.mark.parametrize(
    "temp, overwrite",
    [
        param(
            True,
            True,
            id="temp overwrite",
            marks=[
                pytest.mark.notyet(["clickhouse"], reason="Can't specify both"),
                pytest.mark.notyet(
                    [
                        "pyspark",
                        "trino",
                        "exasol",
                        "risingwave",
                        "impala",
                        "databricks",
                    ],
                    reason="No support for temp tables",
                ),
                pytest.mark.notyet(
                    ["mssql"],
                    reason="Can't rename temp tables",
                    raises=ValueError,
                ),
                pytest.mark.notimpl(
                    ["bigquery"],
                    reason="tables created with temp=True cause a 404 on retrieval",
                ),
            ],
        ),
        param(
            False,
            True,
            id="no temp, overwrite",
            marks=pytest.mark.notyet(["flink", "polars"]),
        ),
        param(
            True,
            False,
            id="temp, no overwrite",
            marks=[
                pytest.mark.notyet(
                    [
                        "pyspark",
                        "trino",
                        "exasol",
                        "risingwave",
                        "impala",
                        "databricks",
                    ],
                    reason="No support for temp tables",
                ),
                pytest.mark.notimpl(["mssql"], reason="Incorrect temp table syntax"),
                pytest.mark.notimpl(
                    ["bigquery"],
                    reason="tables created with temp=True cause a 404 on retrieval",
                ),
            ],
        ),
    ],
)
@pytest.mark.notimpl(["druid"])
def test_create_table_overwrite_temp(backend, con, temp_table, temp, overwrite):
    df = pd.DataFrame(
        {
            "first_name": ["A", "B", "C"],
            "last_name": ["D", "E", "F"],
            "department_name": ["AA", "BB", "CC"],
            "salary": [100.0, 200.0, 300.0],
        }
    )

    con.create_table(temp_table, df, temp=temp, overwrite=overwrite)
    if overwrite:
        con.create_table(temp_table, df, temp=temp, overwrite=overwrite)
    result = (
        con.table(temp_table).execute().sort_values("first_name").reset_index(drop=True)
    )

    backend.assert_frame_equal(df, result)


@pytest.mark.parametrize(
    "lamduh",
    [(lambda df: df), (lambda df: pa.Table.from_pandas(df))],
    ids=["dataframe", "pyarrow table"],
)
@pytest.mark.notyet(["druid"], raises=NotImplementedError)
@pytest.mark.notyet(["flink"], raises=com.IbisError)
def test_load_data(backend, con, temp_table, lamduh):
    sch = ibis.schema(
        [
            ("first_name", "string"),
            ("last_name", "string"),
            ("department_name", "string"),
            ("salary", "float64"),
        ]
    )

    df = pd.DataFrame(
        {
            "first_name": ["A", "B", "C"],
            "last_name": ["D", "E", "F"],
            "department_name": ["AA", "BB", "CC"],
            "salary": [100.0, 200.0, 300.0],
        }
    )

    obj = lamduh(df)
    con.create_table(temp_table, obj, schema=sch, overwrite=True)
    result = (
        con.table(temp_table).execute().sort_values("first_name").reset_index(drop=True)
    )

    backend.assert_frame_equal(df, result)


@mark.parametrize(
    ("expr_fn", "expected"),
    [
        param(lambda t: t.string_col, [("string_col", dt.String)], id="column"),
        param(
            lambda t: t.select(t.string_col, t.bigint_col),
            [("string_col", dt.String), ("bigint_col", dt.Int64)],
            id="table",
        ),
    ],
)
def test_query_schema(ddl_backend, expr_fn, expected):
    expr = expr_fn(ddl_backend.functional_alltypes)

    # we might need a public API for it
    schema = expr.as_table().schema()

    # clickhouse columns has been defined as non-nullable
    # whereas other backends don't support non-nullable columns yet
    expected = ibis.schema(
        [(name, dtype(nullable=schema[name].nullable)) for name, dtype in expected]
    )
    assert schema.equals(expected)


_LIMIT = {
    "oracle": "FETCH FIRST 10 ROWS ONLY",
}


@pytest.mark.notimpl(["mssql"])
def test_sql(backend, con):
    # execute the expression using SQL query
    table = backend.format_table("functional_alltypes")
    limit = _LIMIT.get(backend.name(), "LIMIT 10")
    expr = con.sql(f"SELECT * FROM {table} {limit}")
    result = expr.execute()
    assert len(result) == 10


backend_type_mapping = {
    # backends only implement int64
    "bigquery": {
        dt.int32: dt.int64,
    },
    "snowflake": {
        dt.int32: dt.int64,
    },
}


@mark.notimpl(["druid"])
@pytest.mark.notimpl(
    ["flink"],
    raises=com.IbisError,
    reason="`tbl_properties` is required when creating table with schema",
)
def test_create_table_from_schema(con, new_schema, temp_table):
    new_table = con.create_table(temp_table, schema=new_schema)
    backend_mapping = backend_type_mapping.get(con.name, {})

    result = ibis.schema(
        {
            column_name: backend_mapping.get(
                new_schema[column_name], new_schema[column_name]
            )
            for column_name in new_table.schema().keys()
        }
    )
    assert result == new_table.schema()


@mark.notimpl(
    ["oracle"],
    reason="oracle temp tables aren't cleaned up on reconnect -- they need to "
    "be switched from using atexit to weakref.finalize",
)
@mark.notimpl(["trino", "druid"], reason="doesn't implement temporary tables")
@mark.notimpl(["exasol"], reason="Exasol does not support temporary tables")
@pytest.mark.notimpl(
    ["impala", "pyspark"],
    reason="temporary tables not implemented",
    raises=NotImplementedError,
)
@pytest.mark.never(
    ["risingwave", "databricks"],
    raises=com.UnsupportedOperationError,
    reason="Feature is not yet implemented: CREATE TEMPORARY TABLE",
)
@pytest.mark.notyet(
    ["datafusion"],
    raises=Exception,
    reason="temp tables are not supported upstream in datafusion",
)
@pytest.mark.notimpl(
    ["flink"],
    raises=com.IbisError,
    reason="`tbl_properties` is required when creating table with schema",
)
def test_create_temporary_table_from_schema(con_no_data, new_schema):
    if con_no_data.name == "snowflake" and os.environ.get("SNOWFLAKE_SNOWPARK"):
        with pytest.raises(
            com.IbisError, match="Cannot reconnect to unconfigured .+ backend"
        ):
            con_no_data.reconnect()
        return

    temp_table = gen_name(f"test_{con_no_data.name}_tmp")
    table = con_no_data.create_table(temp_table, schema=new_schema, temp=True)

    # verify table exist in the current session
    backend_mapping = backend_type_mapping.get(con_no_data.name, dict())
    for column_name, column_type in table.schema().items():
        assert (
            backend_mapping.get(new_schema[column_name], new_schema[column_name])
            == column_type
        )

    con_no_data.reconnect()
    # verify table no longer exist after reconnect
    assert temp_table not in con_no_data.list_tables()


@mark.notimpl(
    [
        "bigquery",
        "clickhouse",
        "datafusion",
        "druid",
        "duckdb",
        "exasol",
        "mssql",
        "mysql",
        "oracle",
        "polars",
        "postgres",
        "risingwave",
        "snowflake",
        "sqlite",
        "trino",
    ]
)
@pytest.mark.notimpl(
    ["flink"],
    raises=com.IbisError,
    reason="`tbl_properties` is required when creating table with schema",
)
def test_rename_table(con, temp_table, temp_table_orig):
    schema = ibis.schema({"a": "string", "b": "bool", "c": "int32"})
    con.create_table(temp_table_orig, schema=schema)
    con.rename_table(temp_table_orig, temp_table)
    new = con.table(temp_table)
    assert new.schema().equals(schema)
    assert temp_table_orig not in con.list_tables()


@mark.notimpl(["polars", "druid"])
@mark.never(["impala", "pyspark", "databricks"], reason="No non-nullable datatypes")
@pytest.mark.notimpl(
    ["flink"],
    raises=com.IbisError,
    reason="`tbl_properties` is required when creating table with schema",
)
@pytest.mark.never(
    ["risingwave"],
    raises=PsycoPg2InternalError,
    reason='Feature is not yet implemented: column constraints "NOT NULL"',
)
def test_nullable_input_output(con, temp_table):
    sch = ibis.schema(
        [("foo", "int64"), ("bar", dt.int64(nullable=False)), ("baz", "boolean")]
    )
    t = con.create_table(temp_table, schema=sch)

    assert t.schema().types[0].nullable
    assert not t.schema().types[1].nullable
    assert t.schema().types[2].nullable


@mark.notimpl(["druid"])
def test_create_drop_view(ddl_con, temp_view):
    # setup
    table_name = "functional_alltypes"
    tables = ddl_con.list_tables()

    if table_name in tables or (table_name := table_name.upper()) in tables:
        expr = ddl_con.table(table_name)
    else:
        raise ValueError(f"table `{table_name}` does not exist")

    expr = expr.limit(1)

    # create a new view
    ddl_con.create_view(temp_view, expr)
    # check if the view was created
    assert temp_view in ddl_con.list_tables()

    t_expr = ddl_con.table(table_name)
    v_expr = ddl_con.table(temp_view)
    # check if the view and the table has the same fields
    assert set(t_expr.schema().names) == set(v_expr.schema().names)


@pytest.fixture
def test_employee_schema() -> ibis.schema:
    return ibis.schema(
        {
            "first_name": "string",
            "last_name": "string",
            "department_name": "string",
            "salary": "float64",
        }
    )


@pytest.fixture
def employee_empty_temp_table(backend, con, test_employee_schema):
    temp_table_name = gen_name("temp_employee_empty_table")
    _create_temp_table_with_schema(backend, con, temp_table_name, test_employee_schema)
    yield temp_table_name
    con.drop_table(temp_table_name, force=True)


@pytest.fixture
def employee_data_1_temp_table(backend, con, test_employee_schema):
    test_employee_data_1 = pd.DataFrame(
        {
            "first_name": ["A", "B", "C"],
            "last_name": ["D", "E", "F"],
            "department_name": ["AA", "BB", "CC"],
            "salary": [100.0, 200.0, 300.0],
        }
    )

    temp_table_name = gen_name("temp_employee_data_1")
    _create_temp_table_with_schema(
        backend, con, temp_table_name, test_employee_schema, data=test_employee_data_1
    )
    assert temp_table_name in con.list_tables()
    yield temp_table_name
    con.drop_table(temp_table_name, force=True)


@pytest.fixture
def test_employee_data_2():
    import pandas as pd

    df2 = pd.DataFrame(
        {
            "first_name": ["X", "Y", "Z"],
            "last_name": ["A", "B", "C"],
            "department_name": ["XX", "YY", "ZZ"],
            "salary": [400.0, 500.0, 600.0],
        }
    )

    return df2


@pytest.fixture
def employee_data_2_temp_table(
    backend, con, test_employee_schema, test_employee_data_2
):
    temp_table_name = gen_name("temp_employee_data_2")
    _create_temp_table_with_schema(
        backend, con, temp_table_name, test_employee_schema, data=test_employee_data_2
    )
    yield temp_table_name
    con.drop_table(temp_table_name, force=True)


@pytest.mark.notimpl(["polars"], reason="`insert` method not implemented")
def test_insert_no_overwrite_from_dataframe(
    backend, con, test_employee_data_2, employee_empty_temp_table
):
    temporary = con.table(employee_empty_temp_table)
    con.insert(employee_empty_temp_table, obj=test_employee_data_2, overwrite=False)
    result = temporary.execute()
    assert len(result) == 3
    backend.assert_frame_equal(
        result.sort_values("first_name").reset_index(drop=True),
        test_employee_data_2.sort_values("first_name").reset_index(drop=True),
    )


@pytest.mark.notimpl(["polars"], reason="`insert` method not implemented")
@pytest.mark.notyet(
    ["risingwave"],
    raises=PsycoPg2InternalError,
    reason="truncate not supported upstream",
)
@pytest.mark.notyet(
    ["datafusion"], raises=Exception, reason="DELETE DML not implemented upstream"
)
@pytest.mark.notyet(["druid"], raises=NotImplementedError)
def test_insert_overwrite_from_dataframe(
    backend, con, employee_data_1_temp_table, test_employee_data_2
):
    temporary = con.table(employee_data_1_temp_table)

    con.insert(employee_data_1_temp_table, obj=test_employee_data_2, overwrite=True)
    result = temporary.execute()
    assert len(result) == 3
    backend.assert_frame_equal(
        result.sort_values("first_name").reset_index(drop=True),
        test_employee_data_2.sort_values("first_name").reset_index(drop=True),
    )


@pytest.mark.notimpl(["polars"], reason="`insert` method not implemented")
def test_insert_no_overwrite_from_expr(
    backend, con, employee_empty_temp_table, employee_data_2_temp_table
):
    temporary = con.table(employee_empty_temp_table)
    from_table = con.table(employee_data_2_temp_table)

    con.insert(employee_empty_temp_table, obj=from_table, overwrite=False)
    result = temporary.execute()
    assert len(result) == 3
    backend.assert_frame_equal(
        result.sort_values("first_name").reset_index(drop=True),
        from_table.execute().sort_values("first_name").reset_index(drop=True),
    )


@pytest.mark.notimpl(["polars"], reason="`insert` method not implemented")
@pytest.mark.notyet(
    ["datafusion"], raises=Exception, reason="DELETE DML not implemented upstream"
)
@pytest.mark.notyet(
    ["risingwave"],
    raises=PsycoPg2InternalError,
    reason="truncate not supported upstream",
)
def test_insert_overwrite_from_expr(
    backend, con, employee_data_1_temp_table, employee_data_2_temp_table
):
    temporary = con.table(employee_data_1_temp_table)
    from_table = con.table(employee_data_2_temp_table)

    con.insert(employee_data_1_temp_table, obj=from_table, overwrite=True)
    result = temporary.execute()
    assert len(result) == 3
    backend.assert_frame_equal(
        result.sort_values("first_name").reset_index(drop=True),
        from_table.execute().sort_values("first_name").reset_index(drop=True),
    )


@pytest.mark.notimpl(["polars"], reason="`insert` method not implemented")
@pytest.mark.notyet(
    ["datafusion"], raises=Exception, reason="DELETE DML not implemented upstream"
)
@pytest.mark.notyet(
    ["risingwave"],
    raises=PsycoPg2InternalError,
    reason="truncate not supported upstream",
)
def test_insert_overwrite_from_list(con, employee_data_1_temp_table):
    def _emp(a, b, c, d):
        return dict(first_name=a, last_name=b, department_name=c, salary=d)

    con.insert(
        employee_data_1_temp_table,
        [
            _emp("Adam", "Smith", "Accounting", 50000.0),
            _emp("Mohammed", "Ali", "Boxing", 150000),
            _emp("María", "Gonzalez", "Engineering", 100000.0),
        ],
        overwrite=True,
    )

    assert len(con.table(employee_data_1_temp_table).execute()) == 3


@pytest.mark.notimpl(
    ["polars"], raises=AttributeError, reason="`insert` method not implemented"
)
@pytest.mark.notyet(["druid"], raises=NotImplementedError)
@pytest.mark.notimpl(["polars"], raises=AttributeError)
@pytest.mark.notimpl(
    ["flink"],
    raises=com.IbisError,
    reason="`tbl_properties` is required when creating table with schema",
)
def test_insert_from_memtable(con, temp_table):
    df = pd.DataFrame({"x": range(3)})
    table_name = temp_table
    mt = ibis.memtable(df)
    con.create_table(table_name, schema=mt.schema())
    con.insert(table_name, mt)
    con.insert(table_name, mt)

    table = con.tables[table_name]
    assert len(table.execute()) == 6
    assert con.tables[table_name].schema() == ibis.schema({"x": "int64"})


@pytest.mark.notyet(
    [
        "bigquery",
        "clickhouse",
        "druid",
        "exasol",
        "impala",
        "mysql",
        "oracle",
        "polars",
        "flink",
        "sqlite",
    ],
    raises=AttributeError,
    reason="doesn't support the common notion of a catalog",
)
@pytest.mark.xfail_version(pyspark=["pyspark<3.4"])
def test_list_catalogs(con):
    # Every backend has its own databases
    test_catalogs = {
        "datafusion": {"datafusion"},
        "duckdb": {"memory"},
        "exasol": set(),
        "flink": set(),
        "mssql": {"ibis_testing"},
        "oracle": set(),
        "postgres": {"postgres", "ibis_testing"},
        "risingwave": {"dev"},
        "snowflake": {"IBIS_TESTING"},
        "trino": {"memory"},
        "pyspark": {"spark_catalog"},
        "databricks": {"hive_metastore", "ibis", "ibis_testing", "samples", "system"},
    }
    result = set(con.list_catalogs())
    assert test_catalogs[con.name] <= result


@pytest.mark.notyet(
    ["druid", "polars"],
    raises=AttributeError,
    reason="doesn't support the common notion of a database",
)
def test_list_database_contents(con):
    # Every backend has its own databases
    test_databases = {
        "bigquery": {"ibis_gbq_testing"},
        "clickhouse": {"system", "default", "ibis_testing"},
        "datafusion": {"public"},
        "duckdb": {"pg_catalog", "main", "information_schema"},
        "exasol": {"EXASOL"},
        "flink": {"default_database"},
        "impala": {"ibis_testing", "default", "_impala_builtins"},
        "mssql": {"INFORMATION_SCHEMA", "dbo", "guest"},
        "mysql": {"ibis-testing", "information_schema"},
        "oracle": {"SYS", "IBIS"},
        "postgres": {"public", "information_schema"},
        "pyspark": set(),
        "risingwave": {"public", "rw_catalog", "information_schema"},
        "snowflake": {"IBIS_TESTING"},
        "sqlite": {"main"},
        "trino": {"default", "information_schema"},
        "databricks": {"default"},
    }
    result = set(con.list_databases())
    assert test_databases[con.name] <= result


@pytest.mark.notyet(["mssql"], raises=PyODBCProgrammingError)
@pytest.mark.notyet(["pyspark"], raises=com.IbisTypeError)
@pytest.mark.notyet(["databricks"], raises=DatabricksServerOperationError)
@pytest.mark.notyet(["bigquery"], raises=com.UnsupportedBackendType)
@pytest.mark.notyet(
    ["postgres"], raises=PsycoPg2UndefinedObject, reason="no unsigned int types"
)
@pytest.mark.notyet(
    ["oracle"], raises=OracleDatabaseError, reason="no unsigned int types"
)
@pytest.mark.notyet(["exasol"], raises=ExaQueryError, reason="no unsigned int types")
@pytest.mark.notyet(["datafusion"], raises=Exception, reason="no unsigned int types")
@pytest.mark.notyet(["druid"], raises=NotImplementedError)
@pytest.mark.notyet(["snowflake"], raises=SnowflakeProgrammingError)
@pytest.mark.notyet(["impala"], raises=ImpalaHiveServer2Error)
@pytest.mark.notyet(
    ["risingwave"],
    raises=PsycoPg2InternalError,
    reason="unsigned integers are not supported",
)
@pytest.mark.notimpl(
    ["flink"],
    raises=com.IbisError,
    reason="`tbl_properties` is required when creating table with schema",
)
def test_unsigned_integer_type(con, temp_table):
    con.create_table(
        temp_table,
        schema=ibis.schema(dict(a="uint8", b="uint16", c="uint32", d="uint64")),
        overwrite=True,
    )
    assert temp_table in con.list_tables()


@pytest.mark.backend
@pytest.mark.parametrize(
    "url",
    [
        param(
            "clickhouse://default@localhost:8123/default",
            marks=mark.clickhouse,
            id="clickhouse",
        ),
        param(
            "datafusion://",
            marks=mark.datafusion,
            id="datafusion",
        ),
        param(
            "impala://localhost:21050/default",
            marks=mark.impala,
            id="impala",
        ),
        param(
            "mysql://ibis:ibis@localhost:3306",
            marks=mark.mysql,
            id="mysql",
        ),
        param(
            "polars://",
            marks=mark.polars,
            id="polars",
        ),
        param(
            "postgres://postgres:postgres@localhost:5432",
            marks=mark.postgres,
            id="postgres",
        ),
        param(
            "postgresql://postgres:postgres@localhost:5432/postgres",
            marks=mark.postgres,
            id="postgresql",
        ),
        param(
            "pyspark://?spark.app.name=test-pyspark",
            marks=[
                mark.pyspark,
                pytest.mark.skipif(
                    sys.version_info >= (3, 11)
                    or vparse(pd.__version__) >= vparse("2")
                    or vparse(np.__version__) >= vparse("1.24"),
                    reason="passes on 3.11, but no other pyspark tests do",
                ),
            ],
            id="pyspark",
        ),
        param(
            "pyspark://my-warehouse-dir?spark.app.name=test-pyspark",
            marks=[
                mark.pyspark,
                pytest.mark.skipif(
                    sys.version_info >= (3, 11)
                    or vparse(pd.__version__) >= vparse("2")
                    or vparse(np.__version__) >= vparse("1.24"),
                    reason="passes on 3.11, but no other pyspark tests do",
                ),
            ],
            id="pyspark_with_warehouse",
        ),
        param(
            "pyspark://my-warehouse-dir",
            marks=[
                mark.pyspark,
                pytest.mark.skipif(
                    sys.version_info >= (3, 11)
                    or vparse(pd.__version__) >= vparse("2")
                    or vparse(np.__version__) >= vparse("1.24"),
                    reason="passes on 3.11, but no other pyspark tests do",
                ),
            ],
            id="pyspark_with_warehouse_no_params",
        ),
    ],
)
def test_connect_url(url):
    con = ibis.connect(url)
    one = ibis.literal(1)
    assert con.execute(one) == 1


@pytest.mark.parametrize(
    ("arg", "lambda_", "expected"),
    [
        param(
            [(1, 2.0, "3")],
            lambda arg: ibis.memtable(arg, columns=list("abc")),
            pd.DataFrame([(1, 2.0, "3")], columns=list("abc")),
            id="simple",
        ),
        param(
            [(1, 2.0, "3")],
            lambda arg: ibis.memtable(arg),
            pd.DataFrame([(1, 2.0, "3")], columns=["col0", "col1", "col2"]),
            id="simple_auto_named",
        ),
        param(
            [(1, 2.0, "3")],
            lambda arg: ibis.memtable(
                arg,
                schema=ibis.schema(dict(a="int8", b="float32", c="string")),
            ),
            pd.DataFrame([(1, 2.0, "3")], columns=list("abc")).astype(
                {"a": "int8", "b": "float32"}
            ),
            id="simple_schema",
        ),
        param(
            pd.DataFrame({"a": [1], "b": [2.0], "c": ["3"]}).astype(
                {"a": "int8", "b": "float32"}
            ),
            lambda arg: ibis.memtable(arg),
            pd.DataFrame([(1, 2.0, "3")], columns=list("abc")).astype(
                {"a": "int8", "b": "float32"}
            ),
            id="dataframe",
        ),
        param(
            [dict(a=1), dict(a=2)],
            lambda arg: ibis.memtable(arg),
            pd.DataFrame({"a": [1, 2]}),
            id="list_of_dicts",
        ),
    ],
)
def test_in_memory_table(backend, con, arg, lambda_, expected, monkeypatch):
    monkeypatch.setattr(ibis.options, "default_backend", con)

    expr = lambda_(arg)
    result = con.execute(expr.order_by(expr.columns[0]))
    backend.assert_frame_equal(result, expected)


def test_filter_memory_table(backend, con, monkeypatch):
    monkeypatch.setattr(ibis.options, "default_backend", con)

    t = ibis.memtable([(1, 2), (3, 4), (5, 6)], columns=["x", "y"])
    expr = t.filter(t.x > 1).order_by("x")
    expected = pd.DataFrame({"x": [3, 5], "y": [4, 6]})
    result = con.execute(expr)
    backend.assert_frame_equal(result, expected)


def test_agg_memory_table(con, monkeypatch):
    monkeypatch.setattr(ibis.options, "default_backend", con)

    t = ibis.memtable([(1, 2), (3, 4), (5, 6)], columns=["x", "y"])
    expr = t.x.count()
    result = con.execute(expr)
    assert result == 3


def test_self_join_memory_table(backend, con, monkeypatch):
    monkeypatch.setattr(ibis.options, "default_backend", con)

    t = ibis.memtable({"x": [1, 2], "y": [2, 1], "z": ["a", "b"]})
    t_view = t.view()
    expr = t.join(t_view, t.x == t_view.y).select("x", "y", "z", "z_right")
    result = con.execute(expr).sort_values("x").reset_index(drop=True)
    expected = pd.DataFrame(
        {"x": [1, 2], "y": [2, 1], "z": ["a", "b"], "z_right": ["b", "a"]}
    )
    backend.assert_frame_equal(result, expected)


@pytest.mark.parametrize(
    "obj, table_name",
    [
        param(lambda: pa.table({"a": ["a"], "b": [1]}), "df_arrow", id="pyarrow table"),
        param(
            lambda: pa.table({"a": ["a"], "b": [1]}).to_reader(),
            "df_arrow_batch_reader",
            marks=[
                pytest.mark.notimpl(
                    [
                        "bigquery",
                        "clickhouse",
                        "duckdb",
                        "exasol",
                        "impala",
                        "mssql",
                        "mysql",
                        "oracle",
                        "postgres",
                        "pyspark",
                        "risingwave",
                        "snowflake",
                        "sqlite",
                        "trino",
                        "databricks",
                    ]
                )
            ],
            id="pyarrow_rbr",
        ),
        param(
            lambda: pa.table({"a": ["a"], "b": [1]}).to_batches()[0],
            "df_arrow_single_batch",
            id="pyarrow_single_batch",
        ),
        param(
            lambda: ds.dataset(pa.table({"a": ["a"], "b": [1]})),
            "df_arrow_dataset",
            marks=[
                pytest.mark.notimpl(
                    [
                        "bigquery",
                        "clickhouse",
                        "exasol",
                        "impala",
                        "mssql",
                        "mysql",
                        "oracle",
                        "postgres",
                        "pyspark",
                        "risingwave",
                        "snowflake",
                        "sqlite",
                        "trino",
                        "databricks",
                    ],
                    raises=com.UnsupportedOperationError,
                    reason="we don't materialize datasets to avoid perf footguns",
                ),
                pytest.mark.notimpl(["polars"], raises=NotImplementedError),
            ],
            id="pyarrow dataset",
        ),
        param(lambda: pd.DataFrame({"a": ["a"], "b": [1]}), "df_pandas", id="pandas"),
        param(
            lambda: pytest.importorskip("polars").DataFrame({"a": ["a"], "b": [1]}),
            "df_polars_eager",
            id="polars dataframe",
        ),
        param(
            lambda: pytest.importorskip("polars").LazyFrame({"a": ["a"], "b": [1]}),
            "df_polars_lazy",
            id="polars lazyframe",
        ),
        param(
            lambda: ibis.memtable([("a", 1)], columns=["a", "b"]),
            "memtable",
            id="memtable_list",
        ),
        param(
            lambda: ibis.memtable(pa.table({"a": ["a"], "b": [1]})),
            "memtable_pa",
            id="memtable pyarrow",
        ),
        param(
            lambda: ibis.memtable(pd.DataFrame({"a": ["a"], "b": [1]})),
            "memtable_pandas",
            id="memtable pandas",
        ),
        param(
            lambda: ibis.memtable(
                pytest.importorskip("polars").DataFrame({"a": ["a"], "b": [1]})
            ),
            "memtable_polars_eager",
            id="memtable polars dataframe",
        ),
        param(
            lambda: ibis.memtable(
                pytest.importorskip("polars").LazyFrame({"a": ["a"], "b": [1]})
            ),
            "memtable_polars_lazy",
            id="memtable polars lazyframe",
        ),
    ],
)
@pytest.mark.notimpl(["druid"])
@pytest.mark.notimpl(
    ["flink"],
    reason="Flink backend supports creating only TEMPORARY VIEW for in-memory data.",
)
def test_create_table_in_memory(con, obj, table_name, monkeypatch):
    monkeypatch.setattr(ibis.options, "default_backend", con)
    table_name = gen_name(table_name)
    t = con.create_table(table_name, obj())

    try:
        assert table_name in con.list_tables()
        assert pa.table({"a": ["a"], "b": [1]}).equals(t.to_pyarrow())
    finally:
        con.drop_table(table_name, force=True)


def test_default_backend_option(con, monkeypatch):
    # verify that there's nothing already set
    assert ibis.options.default_backend is None

    monkeypatch.setattr(ibis.options, "default_backend", con)

    backend = ibis.config._default_backend()
    assert backend.name == con.name


# backend is used to ensure that this test runs in CI in the setting
# where only the dependencies for a given backend are installed
@pytest.mark.usefixtures("backend")
def test_default_backend_no_duckdb():
    script = """\
import sys
sys.modules["duckdb"] = None

import ibis

t = ibis.memtable([{'a': 1}, {'a': 2}, {'a': 3}])
t.execute()"""

    args = [sys.executable, "-c", script]
    with pytest.raises(subprocess.CalledProcessError) as e:
        subprocess.check_output(args, stderr=subprocess.STDOUT, universal_newlines=True)
    assert (
        re.search(
            "You have used a function that relies on the default backend",
            e.value.output,
        )
        is not None
    )


@pytest.mark.usefixtures("backend")
def test_default_backend_no_duckdb_read_parquet():
    script = """\
import sys
sys.modules["duckdb"] = None

import ibis
ibis.read_parquet("foo.parquet")"""

    args = [sys.executable, "-c", script]
    with pytest.raises(subprocess.CalledProcessError) as e:
        subprocess.check_output(args, stderr=subprocess.STDOUT, universal_newlines=True)
    assert (
        re.search(
            "You have used a function that relies on the default backend",
            e.value.output,
        )
        is not None
    )


@pytest.mark.parametrize("dtype", [None, "f8"])
def test_dunder_array_table(alltypes, dtype):
    expr = alltypes.group_by("string_col").int_col.sum().order_by("string_col")
    result = np.asarray(expr, dtype=dtype)
    expected = np.asarray(expr.execute(), dtype=dtype)
    np.testing.assert_array_equal(result, expected)


@pytest.mark.parametrize("dtype", [None, "f8"])
def test_dunder_array_column(alltypes, dtype):
    from ibis import _

    expr = alltypes.group_by("string_col").agg(int_col=_.int_col.sum()).int_col
    result = np.sort(np.asarray(expr, dtype=dtype))
    expected = np.sort(np.asarray(expr.execute(), dtype=dtype))
    np.testing.assert_array_equal(result, expected)


@pytest.mark.parametrize("interactive", [True, False])
def test_repr(alltypes, interactive, monkeypatch):
    pytest.importorskip("rich")

    monkeypatch.setattr(ibis.options, "interactive", interactive)

    expr = alltypes.select("date_string_col")

    s = repr(expr)
    # no control characters
    assert all(c.isprintable() or c in "\n\r\t" for c in s)
    if interactive:
        assert "/" in s
    else:
        assert "/" not in s


@pytest.mark.parametrize("show_types", [True, False])
def test_interactive_repr_show_types(alltypes, show_types, monkeypatch):
    pytest.importorskip("rich")

    monkeypatch.setattr(ibis.options, "interactive", True)
    monkeypatch.setattr(ibis.options.repr.interactive, "show_types", show_types)

    expr = alltypes.select("id")
    s = repr(expr)
    if show_types:
        assert "int" in s
    else:
        assert "int" not in s


@pytest.mark.parametrize("is_jupyter", [True, False])
def test_interactive_repr_max_columns(alltypes, is_jupyter, monkeypatch):
    pytest.importorskip("rich")

    monkeypatch.setattr(ibis.options, "interactive", True)

    cols = {f"c_{i}": ibis._.id + i for i in range(50)}
    expr = alltypes.mutate(**cols).select(*cols)

    console = rich.console.Console(force_jupyter=is_jupyter, width=80)
    options = console.options.copy()

    # max_columns = 0
    text = "".join(s.text for s in console.render(expr, options))
    assert " c_0 " in text
    if is_jupyter:
        # All columns are written
        assert " c_49 " in text
    else:
        # width calculations truncates well before 20 columns
        assert " c_19 " not in text

    # max_columns = 3
    monkeypatch.setattr(ibis.options.repr.interactive, "max_columns", 3)
    text = "".join(s.text for s in console.render(expr, options))
    assert " c_2 " in text
    assert " c_3 " not in text

    # max_columns = None
    monkeypatch.setattr(ibis.options.repr.interactive, "max_columns", None)
    text = "".join(s.text for s in console.render(expr, options))
    assert " c_0 " in text
    if is_jupyter:
        # All columns written
        assert " c_49 " in text
    else:
        # width calculations still truncates
        assert " c_19 " not in text


@pytest.mark.parametrize("expr_type", ["table", "column"])
@pytest.mark.parametrize("interactive", [True, False])
def test_repr_mimebundle(alltypes, interactive, expr_type, monkeypatch):
    pytest.importorskip("rich")

    monkeypatch.setattr(ibis.options, "interactive", interactive)

    if expr_type == "column":
        expr = alltypes.date_string_col
    else:
        expr = alltypes.select("date_string_col")

    reprs = expr._repr_mimebundle_(include=["text/plain", "text/html"], exclude=[])
    for format in ["text/plain", "text/html"]:
        if interactive:
            assert "r0.date_string_col" not in reprs[format]
        else:
            assert "r0.date_string_col" in reprs[format]


@pytest.mark.never(
    ["postgres", "bigquery", "duckdb"],
    reason="These backends explicitly do support Geo operations",
)
@pytest.mark.parametrize("op", [ops.GeoDistance, ops.GeoAsText, ops.GeoUnaryUnion])
def test_has_operation_no_geo(con, op):
    """Previously some backends mistakenly reported Geo operations as
    supported.

    Since most backends don't support Geo operations, we test that
    they're excluded here, skipping the few backends that explicitly do
    support them.
    """
    assert not con.has_operation(op)


@pytest.mark.parametrize(
    ("module_name", "op"),
    [
        param(backend, obj, marks=getattr(mark, backend), id=f"{backend}-{name}")
        for name, obj in sorted(inspect.getmembers(builtins), key=itemgetter(0))
        for backend in sorted(ALL_BACKENDS)
        # filter out builtins that are types, except for tuples on ClickHouse
        # and duckdb because tuples are used to represent lists of expressions
        if isinstance(obj, type)
        if (obj is not tuple or backend not in ("clickhouse", "duckdb"))
        if (backend != "pyspark" or vparse(pd.__version__) < vparse("2"))
    ],
)
def test_has_operation_no_builtins(module_name, op):
    mod = importlib.import_module(f"ibis.backends.{module_name}")
    assert not mod.Backend.has_operation(op)


def test_get_backend(con, alltypes, monkeypatch):
    assert ibis.get_backend(alltypes) is con
    assert ibis.get_backend(alltypes.id.min()) is con

    with pytest.raises(com.IbisError, match="contains unbound tables"):
        ibis.get_backend(ibis.table({"x": "int"}))

    monkeypatch.setattr(ibis.options, "default_backend", con)
    assert ibis.get_backend() is con
    expr = ibis.literal(1) + 2
    assert ibis.get_backend(expr) is con


def test_set_backend(con, monkeypatch):
    monkeypatch.setattr(ibis.options, "default_backend", None)
    ibis.set_backend(con)
    assert ibis.get_backend() is con


@pytest.mark.parametrize(
    "name",
    [
        param(name, marks=getattr(mark, name), id=name)
        for name in ("datafusion", "duckdb", "polars", "sqlite")
    ],
)
def test_set_backend_name(name, monkeypatch):
    # Don't need to test with all backends, only checking that things are
    # plumbed through correctly.
    monkeypatch.setattr(ibis.options, "default_backend", None)
    ibis.set_backend(name)
    assert ibis.get_backend().name == name


@pytest.mark.parametrize(
    "url",
    [
        param(
            "clickhouse://default@localhost:8123",
            marks=mark.clickhouse,
            id="clickhouse",
        ),
        param(
            "mysql://ibis:ibis@localhost:3306",
            marks=mark.mysql,
            id="mysql",
        ),
        param(
            "postgres://postgres:postgres@localhost:5432",
            marks=mark.postgres,
            id="postgres",
        ),
    ],
)
def test_set_backend_url(url, monkeypatch):
    # Don't need to test with all backends, only checking that things are
    # plumbed through correctly.
    monkeypatch.setattr(ibis.options, "default_backend", None)
    name = url.split("://")[0]
    ibis.set_backend(url)
    assert ibis.get_backend().name == name


@pytest.mark.notyet(
    [
        "bigquery",
        "datafusion",
        "duckdb",
        "exasol",
        "impala",
        "mssql",
        "mysql",
        "polars",
        "postgres",
        "risingwave",
        "pyspark",
        "sqlite",
        "databricks",
    ],
    reason="backend doesn't support timestamp with scale parameter",
)
@pytest.mark.notimpl(["clickhouse"], reason="create table isn't implemented")
@pytest.mark.notimpl(
    ["snowflake"], reason="scale not implemented in ibis's snowflake backend"
)
@pytest.mark.never(
    ["oracle"], reason="oracle doesn't allow DESCRIBE outside of its CLI"
)
@pytest.mark.notimpl(["druid"], reason="dialect is broken")
@pytest.mark.notimpl(
    ["flink"],
    raises=com.IbisError,
    reason="`tbl_properties` is required when creating table with schema",
)
def test_create_table_timestamp(con, temp_table):
    schema = ibis.schema(
        dict(zip(string.ascii_letters, map("timestamp({:d})".format, range(10))))
    )
    con.create_table(
        temp_table,
        schema=schema,
        overwrite=True,
    )
    rows = con.raw_sql(f"DESCRIBE {temp_table}").fetchall()
    result = ibis.schema((name, typ) for name, typ, *_ in rows)
    assert result == schema


@contextlib.contextmanager
def gen_test_name(con: BaseBackend):
    name = gen_name("test_table")
    yield name
    con.drop_table(name, force=True)


@mark.notimpl(
    ["druid"], raises=NotImplementedError, reason="generated SQL fails to parse"
)
@mark.notimpl(["impala"], reason="impala doesn't support memtable")
@mark.notimpl(["pyspark"])
@mark.notimpl(
    ["flink"],
    raises=com.IbisError,
    reason=(
        "Unsupported operation: <class 'ibis.expr.operations.relations.Selection'>. "
        "If `obj` is of `ir.Table`, the operation must be `InMemoryTable`."
    ),
)
def test_overwrite(ddl_con, monkeypatch):
    monkeypatch.setattr(ibis.options, "default_backend", ddl_con)

    t0 = ibis.memtable({"a": [1, 2, 3]})

    with gen_test_name(ddl_con) as x:
        t1 = ddl_con.create_table(x, t0)
        t2 = t1.filter(t1.a < 3)
        expected_count = 2

        assert t2.count().execute() == expected_count

        with gen_test_name(ddl_con) as y:
            assert ddl_con.create_table(y, t2).count().execute() == expected_count

        assert (
            ddl_con.create_table(x, t2, overwrite=True).count().execute()
            == expected_count
        )
        assert t2.count().execute() == expected_count


@contextlib.contextmanager
def create_and_destroy_db(con):
    con.create_database(dbname := gen_name("db"))
    try:
        yield dbname
    finally:
        con.drop_database(dbname)


# TODO: move this to something like `test_ddl.py`
@pytest.mark.notyet(
    ["flink"],
    reason="unclear whether Flink supports cross catalog/database inserts",
    raises=Py4JJavaError,
)
def test_insert_with_database_specified(con_create_database):
    con = con_create_database

    t = ibis.memtable({"a": [1, 2, 3]})

    with create_and_destroy_db(con) as dbname:
        con.create_table(
            table_name := gen_name("table"),
            obj=t,
            database=dbname,
            temp=con.name == "flink",
        )
        try:
            con.insert(table_name, obj=t, database=dbname)
            assert con.table(table_name, database=dbname).count().to_pandas() == 6
        finally:
            con.drop_table(table_name, database=dbname)


@pytest.mark.notyet(["datafusion"], reason="cannot list or drop catalogs")
def test_create_catalog(con_create_catalog):
    catalog = gen_name("test_create_catalog")
    con_create_catalog.create_catalog(catalog)
    assert catalog in con_create_catalog.list_catalogs()
    con_create_catalog.drop_catalog(catalog)
    assert catalog not in con_create_catalog.list_catalogs()


def test_create_database(con_create_database):
    database = gen_name("test_create_database")
    con_create_database.create_database(database)
    assert database in con_create_database.list_databases()
    con_create_database.drop_database(database)
    assert database not in con_create_database.list_databases()


def test_list_databases(con_create_database):
    databases = con_create_database.list_databases()
    assert len(databases) == len(set(databases))


@pytest.mark.notyet(["datafusion"], reason="cannot list or drop databases")
def test_create_catalog_database(con_create_catalog_database):
    catalog = gen_name("test_create_catalog")
    con_create_catalog_database.create_catalog(catalog)
    try:
        database = gen_name("test_create_catalog_database")
        con_create_catalog_database.create_database(database, catalog=catalog)
        con_create_catalog_database.drop_database(database, catalog=catalog)
    finally:
        con_create_catalog_database.drop_catalog(catalog)


@pytest.mark.notyet(["datafusion"], reason="cannot list or drop databases")
def test_list_catalogs_databases(con_create_catalog_database):
    catalog = gen_name("test_create_catalog")
    con_create_catalog_database.create_catalog(catalog)
    try:
        database = gen_name("test_create_catalog_database")
        con_create_catalog_database.create_database(database, catalog=catalog)

        try:
            assert database in con_create_catalog_database.list_databases(
                catalog=catalog
            )
        finally:
            con_create_catalog_database.drop_database(database, catalog=catalog)
    finally:
        con_create_catalog_database.drop_catalog(catalog)


@pytest.mark.notyet(
    ["polars", "datafusion"], reason="this is a no-op for in-memory backends"
)
@pytest.mark.notyet(
    ["trino", "clickhouse", "impala", "bigquery", "flink"],
    reason="Backend client does not conform to DB-API, subsequent op does not raise",
)
@pytest.mark.skip()
def test_close_connection(con):
    if con.name == "pyspark":
        # It would be great if there were a simple way to say "give me a new
        # spark context" but I haven't found it.
        pytest.skip("Closing spark context breaks subsequent tests")
    new_con = getattr(ibis, con.name).connect(*con._con_args, **con._con_kwargs)

    # Run any command that hits the backend
    _ = new_con.list_tables()
    new_con.disconnect()

    # DB-API states that subsequent execution attempt should raise
    with pytest.raises(Exception):  # noqa:B017
        new_con.list_tables()


@pytest.mark.notyet(
    ["clickhouse"],
    raises=AttributeError,
    reason="JSON extension is experimental and not enabled by default in testing",
)
@pytest.mark.notyet(
    ["datafusion", "polars", "mssql", "druid", "oracle", "exasol", "impala"],
    raises=AttributeError,
    reason="JSON type not implemented",
)
@pytest.mark.notimpl(
    ["risingwave", "sqlite"],
    raises=pa.ArrowTypeError,
    reason="mismatch between output value and expected input type",
)
@pytest.mark.never(
    ["snowflake"],
    raises=TypeError,
    reason="snowflake uses a custom pyarrow extension type for JSON pretty printing",
)
@pytest.mark.notimpl(
    ["databricks"], raises=json.JSONDecodeError, reason="not yet implemented"
)
def test_json_to_pyarrow(con):
    t = con.tables.json_t
    table = t.to_pyarrow()
    js = table["js"]

    expected = [
        {"a": [1, 2, 3, 4], "b": 1},
        {"a": None, "b": 2},
        {"a": "foo", "c": None},
        None,
        [42, 47, 55],
        [],
        "a",
        "",
        "b",
        None,
        True,
        False,
        42,
        37.37,
    ]
    expected = {json.dumps(val) for val in expected}

    result = {
        # loads and dumps so the string representation is the same
        json.dumps(json.loads(val))
        for val in js.to_pylist()
        # proper null values must be ignored because they cannot be
        # deserialized as JSON
        #
        # they exist in the json_t table, so the `js` value contains them
        if val is not None
    }
    assert result == expected


@pytest.mark.notyet(
    ["risingwave", "exasol", "databricks"],
    raises=com.UnsupportedOperationError,
    reason="no temp table support",
)
@pytest.mark.notyet(
    ["impala", "trino"], raises=NotImplementedError, reason="no temp table support"
)
@pytest.mark.notyet(
    ["druid"], raises=NotImplementedError, reason="doesn't support create_table"
)
@pytest.mark.notyet(
    ["flink"], raises=com.IbisError, reason="no persistent temp table support"
)
def test_schema_with_caching(alltypes):
    t1 = alltypes.limit(5).select("bigint_col", "string_col")
    t2 = alltypes.limit(5).select("string_col", "bigint_col")

    pt1 = t1.cache()
    pt2 = t2.cache()

    assert pt1.schema() == t1.schema()
    assert pt2.schema() == t2.schema()


@pytest.mark.notyet(
    ["druid"], raises=NotImplementedError, reason="doesn't support create_table"
)
@pytest.mark.notyet(["polars"], reason="Doesn't support insert")
@pytest.mark.notyet(
    ["datafusion"], reason="Doesn't support table creation from records"
)
@pytest.mark.notimpl(
    ["flink"], reason="Temp tables are implemented as views, which don't support insert"
)
@pytest.mark.parametrize(
    "first_row, second_row",
    [
        param([{"a": 1, "b": 2}], [{"b": 22, "a": 11}], id="column order reversed"),
        param([{"a": 1, "b": 2}], [{"a": 11, "b": 22}], id="column order matching"),
        param(
            [{"a": 1, "b": 2}],
            [(11, 22)],
            marks=[
                pytest.mark.notimpl(
                    ["impala"],
                    reason="Impala DDL has strict validation checks on schema",
                )
            ],
            id="auto generated cols",
        ),
    ],
)
def test_insert_using_col_name_not_position(con, first_row, second_row, monkeypatch):
    monkeypatch.setattr(ibis.options, "default_backend", con)
    table_name = gen_name("table")
    con.create_table(table_name, first_row)
    con.insert(table_name, second_row)

    result = con.table(table_name).order_by("a").to_pyarrow()
    expected_result = pa.table({"a": [1, 11], "b": [2, 22]})

    assert result.equals(expected_result)

    # Ideally we'd use a temp table for this test, but several backends don't
    # support them and it's nice to know that data are being inserted correctly.
    con.drop_table(table_name)


CON_ATTR = {"bigquery": "client", "flink": "_table_env", "pyspark": "_session"}
DEFAULT_CON_ATTR = "con"


@pytest.mark.parametrize("top_level", [True, False])
@pytest.mark.never(["polars"], reason="don't have a connection concept")
def test_from_connection(con, top_level):
    backend = getattr(ibis, con.name) if top_level else type(con)
    new_con = backend.from_connection(getattr(con, CON_ATTR.get(con.name, "con")))
    result = int(new_con.execute(ibis.literal(1, type="int")))
    assert result == 1


def test_table_not_found(con):
    with pytest.raises(com.TableNotFound):
        con.table(gen_name("table_not_found"))


@pytest.mark.notimpl(
    ["flink"], raises=com.IbisError, reason="not yet implemented for Flink"
)
def test_no_accidental_cross_database_table_load(con_create_database):
    con = con_create_database

    # Create an extra database
    con.create_database(dbname := gen_name("dummy_db"))

    # Create table with same name in current db and dummy db
    con.create_table(
        table := gen_name("table"), schema=(sch1 := ibis.schema({"a": "int"}))
    )

    con.create_table(table, schema=ibis.schema({"b": "string"}), database=dbname)

    # Can grab table object from current db:
    t = con.table(table)
    assert t.schema().equals(sch1)

    con.drop_table(table)

    # Now attempting to load same table name without specifying db should fail
    with pytest.raises(com.TableNotFound):
        t = con.table(table)

    # But can load if specify other db
    t = con.table(table, database=dbname)

    # Clean up
    con.drop_table(table, database=dbname)
    con.drop_database(dbname)


@pytest.mark.notyet(["druid"], reason="can't create tables")
@pytest.mark.notyet(
    ["flink"], reason="can't create non-temporary tables from in-memory data"
)
def test_cross_database_join(con_create_database, monkeypatch):
    con = con_create_database

    monkeypatch.setattr(ibis.options, "default_backend", con)

    left = ibis.memtable({"a": [1], "b": [2]})
    right = ibis.memtable({"a": [1], "c": [3]})

    # Create an extra database
    con.create_database(dbname := gen_name("dummy_db"))

    # Insert left into current_database
    left = con.create_table(left_table := gen_name("left"), obj=left)

    # Insert right into new database
    right = con.create_table(
        right_table := gen_name("right"), obj=right, database=dbname
    )

    expr = left.join(right, "a")
    assert expr.columns == ("a", "b", "c")

    result = expr.to_pyarrow()
    expected = pa.Table.from_pydict({"a": [1], "b": [2], "c": [3]})

    assert result.equals(expected)

    con.drop_table(left_table)
    con.drop_table(right_table, database=dbname)
    con.drop_database(dbname)


@pytest.mark.notimpl(
    ["druid"], raises=AttributeError, reason="doesn't implement `raw_sql`"
)
@pytest.mark.notimpl(["clickhouse"], reason="create table isn't implemented")
@pytest.mark.notyet(["flink"], raises=Py4JJavaError)
@pytest.mark.notyet(["polars"], reason="Doesn't support insert")
@pytest.mark.notyet(["exasol"], reason="Backend does not support raw_sql")
@pytest.mark.notimpl(
    ["impala", "pyspark", "trino"], reason="Default constraints are not supported"
)
@pytest.mark.notimpl(
    ["databricks"],
    reason="Default constraints ARE supported, "
    "but you have to enable them with a property AND set DEFAULT, so no",
    raises=DatabricksServerOperationError,
)
def test_insert_into_table_missing_columns(con, temp_table):
    db = getattr(con, "current_database", None)

    raw_ident = sg.table(
        temp_table,
        db=db if db is None else sg.to_identifier(db, quoted=True),
        quoted=True,
    ).sql("duckdb")

    ct_sql = f'CREATE TABLE {raw_ident} ("a" INT DEFAULT 1, "b" INT)'
    sg_expr = sg.parse_one(ct_sql, read="duckdb")
    con.raw_sql(sg_expr.sql(dialect=con.dialect))
    con.insert(temp_table, [{"b": 1}])

    result = con.table(temp_table).to_pyarrow().to_pydict()
    expected_result = {"a": [1], "b": [1]}

    assert result == expected_result


@pytest.mark.notyet(["druid"], raises=AssertionError, reason="can't drop tables")
@pytest.mark.notyet(
    ["clickhouse", "flink"],
    raises=AssertionError,
    reason="memtables are assembled every time",
)
@pytest.mark.notyet(
    ["bigquery"], raises=AssertionError, reason="test is flaky", strict=False
)
@pytest.mark.notyet(
    ["pyspark"],
    condition=IS_SPARK_REMOTE,
    raises=AssertionError,
    reason="likely deadlock when using spark connect",
)
def test_memtable_cleanup(con):
    name = ibis.util.gen_name("temp_memtable")
    t = ibis.memtable({"a": [1, 2, 3], "b": list("def")}, name=name)

    # the table isn't registered until we actually execute, and since we
    # haven't yet executed anything, the table shouldn't be there
    assert name not in con.list_tables()

    # execute, which means the table is registered and should be visible in
    # con.list_tables()
    con.execute(t.select("a"))
    assert name in con.list_tables()

    con.execute(t.select("b"))
    assert name in con.list_tables()

    # remove all references to `t`, which means the `op` shouldn't be reachable
    # and the table should thus be dropped and no longer visible
    del t
    assert name not in con.list_tables()


def test_same_name_memtable_is_overwritten(con):
    name = ibis.util.gen_name("temp_memtable")

    data = {"a": [1, 2, 3], "b": ["a", "b", "c"]}

    t = ibis.memtable(data, name=name)
    assert len(con.execute(t)) == 3

    s = ibis.memtable({"a": [1], "b": ["a"]}, name=name)
    assert len(con.execute(s)) == 1


@pytest.mark.notimpl(
    ["clickhouse", "flink"],
    raises=AssertionError,
    reason="backend doesn't use _register_in_memory_table",
)
def test_memtable_registered_exactly_once(con, mocker):
    name = ibis.util.gen_name("temp_memtable")

    spy = mocker.spy(con, "_register_in_memory_table")

    data = {"a": [1, 2, 3], "b": ["a", "b", "c"]}

    t = ibis.memtable(data, name=name)

    assert len(con.execute(t)) == 3
    assert len(con.execute(t)) == 3

    spy.assert_called_once_with(t.op())


def test_unreachable_memtable_does_not_clobber_existing_data(con):
    name = ibis.util.gen_name("temp_memtable")

    t1 = ibis.memtable({"a": [1, 2, 3]}, name=name)
    t2 = ibis.memtable({"a": [4, 5]}, name=name)

    assert len(con.execute(t1)) == 3

    assert len(con.execute(t2)) == 2

    assert len(con.execute(t1)) == 3

    # Drop t1, its finalizer runs, deleting table `test`
    del t1

    # ensure that the previous clean up doesn't erase t2
    assert len(con.execute(t2)) == 2


def test_identically_named_memtables_cannot_be_joined(con):
    name = ibis.util.gen_name("temp_memtable")

    t1 = ibis.memtable({"a": [1, 2, 3]}, name=name)
    t2 = ibis.memtable({"a": [4, 5]}, name=name)

    # mixing two memtables with the same name is an error
    expr = t1.join(t2, "a")
    with pytest.raises(
        com.IbisError, match=rf"Duplicate in-memory table names: \['{name}'\]"
    ):
        con.execute(expr)



from __future__ import annotations

import pandas as pd
import pandas.testing as tm
import pytest

import ibis
import ibis.common.exceptions as exc


@pytest.fixture(scope="session")
def external_duckdb_file(tmpdir_factory):  # pragma: no cover
    ddb_path = str(tmpdir_factory.mktemp("data") / "starwars.ddb")
    con = ibis.duckdb.connect(ddb_path)

    starwars_df = pd.DataFrame(
        {
            "name": ["Luke Skywalker", "C-3PO", "R2-D2"],
            "height": [172, 167, 96],
            "mass": [77.0, 75.0, 32.0],
        }
    )
    con.create_table("starwars", obj=starwars_df)
    con.disconnect()

    return ddb_path, starwars_df


def test_read_write_external_catalog(con, external_duckdb_file, monkeypatch):
    monkeypatch.setattr(ibis.options, "default_backend", con)

    ddb_path, starwars_df = external_duckdb_file
    con.attach(ddb_path, name="ext")

    # Read from catalog
    assert "ext" in con.list_catalogs()
    assert "main" in con.list_databases(catalog="ext")

    assert "starwars" in con.list_tables(database="ext.main")
    assert "starwars" not in con.list_tables()

    starwars = con.table("starwars", database="ext.main")
    tm.assert_frame_equal(starwars.to_pandas(), starwars_df)

    # Write to catalog
    t = ibis.memtable([{"a": 1, "b": "foo"}, {"a": 2, "b": "baz"}])

    _ = con.create_table("t2", obj=t, database="ext.main")

    assert "t2" in con.list_tables(database="ext.main")
    assert "t2" not in con.list_tables()

    table = con.table("t2", database="ext.main")

    tm.assert_frame_equal(t.to_pandas(), table.to_pandas())

    # Overwrite table in catalog

    t_overwrite = ibis.memtable([{"a": 8, "b": "bing"}, {"a": 9, "b": "bong"}])

    _ = con.create_table("t2", obj=t_overwrite, database="ext.main", overwrite=True)

    assert "t2" in con.list_tables(database="ext.main")
    assert "t2" not in con.list_tables()

    table = con.table("t2", database="ext.main")

    tm.assert_frame_equal(t_overwrite.to_pandas(), table.to_pandas())


def test_raise_if_catalog_and_temp(con):
    with pytest.raises(exc.UnsupportedArgumentError):
        con.create_table("some_table", obj="hi", temp=True, database="ext.main")



from __future__ import annotations

import pytest

import ibis
from ibis import _


@pytest.fixture
def table():
    return ibis.table(
        {"ints": "int", "floats": "float", "bools": "bool", "strings": "string"}
    )


@pytest.mark.parametrize(
    "func",
    [
        pytest.param(lambda t, **kws: t.strings.arbitrary(**kws), id="arbitrary"),
        pytest.param(lambda t, **kws: t.strings.collect(**kws), id="collect"),
        pytest.param(lambda t, **kws: t.strings.group_concat(**kws), id="group_concat"),
        pytest.param(
            lambda t, **kws: t.strings.approx_nunique(**kws), id="approx_nunique"
        ),
        pytest.param(
            lambda t, **kws: t.strings.approx_median(**kws), id="approx_median"
        ),
        pytest.param(lambda t, **kws: t.strings.mode(**kws), id="mode"),
        pytest.param(lambda t, **kws: t.strings.max(**kws), id="max"),
        pytest.param(lambda t, **kws: t.strings.min(**kws), id="min"),
        pytest.param(lambda t, **kws: t.strings.argmax(t.ints, **kws), id="argmax"),
        pytest.param(lambda t, **kws: t.strings.argmin(t.ints, **kws), id="argmin"),
        pytest.param(lambda t, **kws: t.strings.median(**kws), id="median"),
        pytest.param(lambda t, **kws: t.strings.quantile(0.25, **kws), id="quantile"),
        pytest.param(
            lambda t, **kws: t.strings.quantile([0.25, 0.75], **kws),
            id="multi-quantile",
        ),
        pytest.param(lambda t, **kws: t.strings.nunique(**kws), id="nunique"),
        pytest.param(lambda t, **kws: t.strings.count(**kws), id="count"),
        pytest.param(lambda t, **kws: t.strings.first(**kws), id="first"),
        pytest.param(lambda t, **kws: t.strings.last(**kws), id="last"),
        pytest.param(lambda t, **kws: t.ints.std(**kws), id="std"),
        pytest.param(lambda t, **kws: t.ints.var(**kws), id="var"),
        pytest.param(lambda t, **kws: t.ints.mean(**kws), id="mean"),
        pytest.param(lambda t, **kws: t.ints.sum(**kws), id="sum"),
        pytest.param(lambda t, **kws: t.ints.corr(t.floats, **kws), id="corr"),
        pytest.param(lambda t, **kws: t.ints.cov(t.floats, **kws), id="cov"),
        pytest.param(lambda t, **kws: t.ints.bit_and(**kws), id="bit_and"),
        pytest.param(lambda t, **kws: t.ints.bit_xor(**kws), id="bit_xor"),
        pytest.param(lambda t, **kws: t.ints.bit_or(**kws), id="bit_or"),
        pytest.param(lambda t, **kws: t.bools.any(**kws), id="any"),
        pytest.param(lambda t, **kws: t.bools.all(**kws), id="all"),
        pytest.param(lambda t, **kws: t.count(**kws), id="table-count"),
        pytest.param(lambda t, **kws: t.nunique(**kws), id="table-nunique"),
    ],
)
def test_aggregation_where(table, func):
    # No where
    op = func(table).op()
    assert op.where is None

    # Literal where
    op = func(table, where=False).op()
    assert op.where.equals(ibis.literal(False).op())

    # Various ways to spell the same column expression
    r1 = func(table, where=table.bools)
    r2 = func(table, where=_.bools)
    r3 = func(table, where=lambda t: t.bools)
    r4 = func(table, where="bools")
    assert r1.equals(r2)
    assert r1.equals(r3)
    assert r1.equals(r4)
    assert r1.op().where.equals(table.bools.op())



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import pandas as pd
import pytest

import ibis
from ibis.backends.conftest import TEST_TABLES
from ibis.backends.tests.base import BackendTest
from ibis.backends.tests.data import array_types, json_types, struct_types, topk, win

if TYPE_CHECKING:
    from pyflink.table import StreamTableEnvironment

TEST_TABLES["functional_alltypes"] = ibis.schema(
    {
        "id": "int32",
        "bool_col": "boolean",
        "tinyint_col": "int8",
        "smallint_col": "int16",
        "int_col": "int32",
        "bigint_col": "int64",
        "float_col": "float32",
        "double_col": "float64",
        "date_string_col": "string",
        "string_col": "string",
        "timestamp_col": "timestamp(3)",  # overriding the higher level fixture with precision because Flink's
        # watermark must use a field of type TIMESTAMP(p) or TIMESTAMP_LTZ(p), where 'p' is from 0 to 3
        "year": "int32",
        "month": "int32",
    }
)


def get_table_env(
    local_env: bool,
    streaming_mode: bool,
) -> StreamTableEnvironment:
    if local_env:
        from pyflink.table import EnvironmentSettings, TableEnvironment

        env_settings = (
            EnvironmentSettings.in_streaming_mode()
            if streaming_mode
            else EnvironmentSettings.in_batch_mode()
        )
        table_env = TableEnvironment.create(env_settings)

    else:
        import os

        from pyflink.java_gateway import get_gateway
        from pyflink.table import StreamTableEnvironment
        from pyflink.table.table_environment import StreamExecutionEnvironment

        # Connect with Flink remote cluster to run the unit tests
        gateway = get_gateway()
        string_class = gateway.jvm.String
        string_array = gateway.new_array(string_class, 0)
        env_settings = (
            gateway.jvm.org.apache.flink.table.api.EnvironmentSettings.inStreamingMode()
            if streaming_mode
            else gateway.jvm.org.apache.flink.table.api.EnvironmentSettings.inBatchMode()
        )
        stream_env = gateway.jvm.org.apache.flink.streaming.api.environment.StreamExecutionEnvironment
        flink_cluster_addr = os.environ.get("FLINK_REMOTE_CLUSTER_ADDR", "localhost")
        flink_cluster_port = int(os.environ.get("FLINK_REMOTE_CLUSTER_PORT", "8081"))
        j_execution_environment = stream_env.createRemoteEnvironment(
            flink_cluster_addr,
            flink_cluster_port,
            env_settings.getConfiguration(),
            string_array,
        )

        env = StreamExecutionEnvironment(j_execution_environment)
        table_env = StreamTableEnvironment.create(env)

    table_config = table_env.get_config()
    table_config.set("table.local-time-zone", "UTC")

    return table_env


class TestConf(BackendTest):
    force_sort = True
    stateful = False
    supports_map = True
    deps = "pandas", "pyflink"

    @staticmethod
    def connect(*, tmpdir, worker_id, **kw: Any):
        """Flink backend is created in batch mode by default. This is to
        comply with the assumption that the tests under ibis/ibis/backends/tests/
        are for batch (storage or processing) backends.
        """

        table_env = get_table_env(local_env=False, streaming_mode=False)
        return ibis.flink.connect(table_env, **kw)

    def _load_data(self, **_: Any) -> None:
        con = self.connection

        for table_name in TEST_TABLES:
            path = self.data_dir / "parquet" / f"{table_name}.parquet"
            con.create_table(table_name, pd.read_parquet(path), temp=True)

        con.create_table("array_types", array_types, temp=True)
        con.create_table("json_t", json_types, temp=True)
        con.create_table("struct", struct_types, temp=True)
        con.create_table("win", win, temp=True)
        con.create_table(
            "map",
            pd.DataFrame(
                {
                    "idx": [1, 2],
                    "kv": [{"a": 1, "b": 2, "c": 3}, {"d": 4, "e": 5, "f": 6}],
                }
            ),
            schema=ibis.schema({"idx": "int64", "kv": "map<string, int64>"}),
            temp=True,
        )
        con.create_table("topk", topk, temp=True)


class TestConfForStreaming(TestConf):
    @staticmethod
    def connect(*, tmpdir, worker_id, **kw: Any):
        """Flink backend is created in streaming mode here. To be used
        in the tests under ibis/ibis/backends/flink/tests/.
        We only use mini cluster here for simplicity.
        """

        table_env = get_table_env(local_env=True, streaming_mode=True)
        return ibis.flink.connect(table_env, **kw)


@pytest.fixture
def simple_schema():
    return [
        ("a", "int8"),
        ("b", "int16"),
        ("c", "int32"),
        ("d", "int64"),
        ("e", "float32"),
        ("f", "float64"),
        ("g", "string"),
        ("h", "boolean"),
        ("i", "timestamp"),
        ("j", "date"),
        ("k", "time"),
    ]


@pytest.fixture
def simple_table(simple_schema):
    return ibis.table(simple_schema, name="table")


@pytest.fixture(scope="session")
def con(tmp_path_factory, data_dir, worker_id):
    return TestConfForStreaming.load_data(
        data_dir, tmp_path_factory, worker_id
    ).connection


@pytest.fixture
def awards_players_schema():
    return TEST_TABLES["awards_players"]


@pytest.fixture
def functional_alltypes_schema():
    return TEST_TABLES["functional_alltypes"]


@pytest.fixture
def csv_source_configs():
    def generate_csv_configs(csv_file):
        return {
            "connector": "filesystem",
            "path": f"ci/ibis-testing-data/csv/{csv_file}.csv",
            "format": "csv",
            "csv.ignore-parse-errors": "true",
        }

    return generate_csv_configs


@pytest.fixture(scope="session")
def functional_alltypes_no_header(tmpdir_factory, data_dir):
    file = tmpdir_factory.mktemp("data") / "functional_alltypes.csv"
    with (
        open(data_dir / "csv" / "functional_alltypes.csv") as reader,
        open(str(file), mode="w") as writer,
    ):
        reader.readline()  # read the first line and discard it
        for line in reader:
            writer.write(line)
    return file


@pytest.fixture(scope="session", autouse=True)
def functional_alltypes_with_watermark(con, functional_alltypes_no_header):
    # create a streaming table with watermark for testing event-time based ops
    t = con.create_table(
        "functional_alltypes_with_watermark",
        schema=TEST_TABLES["functional_alltypes"],
        tbl_properties={
            "connector": "filesystem",
            "path": functional_alltypes_no_header,
            "format": "csv",
        },
        watermark=ibis.watermark("timestamp_col", ibis.interval(seconds=10)),
        temp=True,
    )
    return t



# Copyright 2015 Cloudera Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import os
import string
from urllib.parse import quote_plus

import hypothesis as h
import hypothesis.strategies as st
import numpy as np
import pandas as pd
import pandas.testing as tm
import pytest
import sqlglot as sg
from pytest import param

import ibis
import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.expr.types as ir
from ibis.backends.tests.errors import PsycoPg2OperationalError
from ibis.util import gen_name

pytest.importorskip("psycopg2")

POSTGRES_TEST_DB = os.environ.get("IBIS_TEST_POSTGRES_DATABASE", "ibis_testing")
IBIS_POSTGRES_HOST = os.environ.get("IBIS_TEST_POSTGRES_HOST", "localhost")
IBIS_POSTGRES_PORT = os.environ.get("IBIS_TEST_POSTGRES_PORT", "5432")
IBIS_POSTGRES_USER = os.environ.get("IBIS_TEST_POSTGRES_USER", "postgres")
IBIS_POSTGRES_PASS = os.environ.get("IBIS_TEST_POSTGRES_PASSWORD", "postgres")


def test_table(alltypes):
    assert isinstance(alltypes, ir.Table)


def test_array_execute(alltypes):
    d = alltypes.limit(10).double_col
    s = d.execute()
    assert isinstance(s, pd.Series)
    assert len(s) == 10


def test_literal_execute(con):
    expr = ibis.literal("1234")
    result = con.execute(expr)
    assert result == "1234"


def test_simple_aggregate_execute(alltypes):
    d = alltypes.double_col.sum()
    v = d.execute()
    assert isinstance(v, float)


def test_list_tables(con):
    assert len(con.list_tables(like="functional")) == 1
    assert {"astronauts", "batting", "diamonds"} <= set(con.list_tables())

    _ = con.create_table("tempy", schema=ibis.schema(dict(id="int")), temp=True)

    assert "tempy" in con.list_tables()
    # temp tables only show up when database='public' (or default)
    assert "tempy" not in con.list_tables(database="tiger")


def test_compile_toplevel(assert_sql):
    t = ibis.table([("foo", "double")], name="t0")

    # it works!
    expr = t.foo.sum()
    assert_sql(expr)


def test_list_catalogs(con):
    assert POSTGRES_TEST_DB is not None
    assert POSTGRES_TEST_DB in con.list_catalogs()


def test_interval_films_schema(con):
    t = con.table("films")
    assert t.len.type() == dt.Interval(unit="m")
    assert issubclass(t.len.execute().dtype.type, np.timedelta64)


@pytest.mark.parametrize(
    ("column", "expected_dtype"),
    [
        # a, b and g are variable length intervals, like YEAR TO MONTH
        param("c", dt.Interval("D"), id="day"),
        param("d", dt.Interval("h"), id="hour"),
        param("e", dt.Interval("m"), id="minute"),
        param("f", dt.Interval("s"), id="second"),
    ],
)
def test_all_interval_types_execute(intervals, column, expected_dtype):
    expr = intervals[column]
    assert expr.type() == expected_dtype

    series = expr.execute()
    assert issubclass(series.dtype.type, np.timedelta64)


def test_unsupported_intervals(con):
    t = con.table("not_supported_intervals")
    assert t["a"].type() == dt.Interval("Y")
    assert t["b"].type() == dt.Interval("Y")
    assert t["g"].type() == dt.Interval("M")


@pytest.mark.parametrize("params", [{}, {"database": "public"}])
def test_create_and_drop_table(con, temp_table, params):
    sch = ibis.schema(
        [
            ("first_name", "string"),
            ("last_name", "string"),
            ("department_name", "string"),
            ("salary", "float64"),
        ]
    )

    t = con.create_table(temp_table, schema=sch, **params)
    assert t is not None
    assert con.table(temp_table, **params) is not None

    con.drop_table(temp_table, **params)

    with pytest.raises(com.TableNotFound, match=temp_table):
        con.table(temp_table, **params)


@pytest.mark.parametrize(
    ("pg_type", "expected_type"),
    [
        param(pg_type, ibis_type, id=pg_type.lower())
        for (pg_type, ibis_type) in [
            ("boolean", dt.boolean),
            ("bytea", dt.binary),
            ("char", dt.string),
            ("bigint", dt.int64),
            ("smallint", dt.int16),
            ("integer", dt.int32),
            ("text", dt.string),
            ("json", dt.json),
            ("point", dt.point),
            ("polygon", dt.polygon),
            ("line", dt.linestring),
            ("real", dt.float32),
            ("double precision", dt.float64),
            ("macaddr", dt.macaddr),
            ("macaddr8", dt.macaddr),
            ("inet", dt.inet),
            ("character", dt.string),
            ("character varying", dt.string),
            ("date", dt.date),
            ("time", dt.time),
            ("time without time zone", dt.time),
            ("timestamp without time zone", dt.Timestamp(scale=6)),
            ("timestamp with time zone", dt.Timestamp("UTC", scale=6)),
            ("interval", dt.Interval("s")),
            ("numeric", dt.decimal),
            ("numeric(3, 2)", dt.Decimal(3, 2)),
            ("uuid", dt.uuid),
            ("jsonb", dt.jsonb),
            ("geometry", dt.geometry),
            ("geography", dt.geography),
        ]
    ],
)
def test_get_schema_from_query(con, pg_type, expected_type):
    name = sg.table(ibis.util.guid()).sql("postgres")
    with con._safe_raw_sql(f"CREATE TEMP TABLE {name} (x {pg_type}, y {pg_type}[])"):
        pass
    expected_schema = ibis.schema(dict(x=expected_type, y=dt.Array(expected_type)))
    result_schema = con._get_schema_using_query(f"SELECT x, y FROM {name}")
    assert result_schema == expected_schema


@pytest.mark.parametrize("col", ["search", "simvec"])
def test_unknown_column_type(con, col):
    awards_players = con.table("awards_players_special_types")
    assert awards_players[col].type().is_unknown()


def test_insert_with_cte(con):
    X = con.create_table("X", schema=ibis.schema(dict(id="int")), temp=True)
    assert "X" in con.list_tables()
    expr = X.join(X.mutate(a=X["id"] + 1), ["id"])
    Y = con.create_table("Y", expr, temp=True)
    assert Y.execute().empty


@pytest.fixture(scope="module")
def contz(con):
    with con.begin() as c:
        c.execute("SHOW TIMEZONE")
        [(tz,)] = c.fetchall()
        c.execute("SET TIMEZONE TO 'America/New_York'")

    yield con

    with con.begin() as c:
        c.execute(f"SET TIMEZONE TO '{tz}'")


def test_timezone_from_column(contz, snapshot):
    with contz.begin() as c:
        c.execute(
            """
            CREATE TEMPORARY TABLE x (
                id BIGINT,
                ts_tz TIMESTAMP WITH TIME ZONE NOT NULL,
                ts_no_tz TIMESTAMP WITHOUT TIME ZONE NOT NULL
            );

            INSERT INTO x VALUES
                (1, '2018-01-01 00:00:01+00', '2018-01-01 00:00:02');

            CREATE TEMPORARY TABLE y AS SELECT 1::BIGINT AS id;
            """
        )

    case = (
        contz.table("x")
        .rename(tz="ts_tz", no_tz="ts_no_tz")
        .left_join(contz.table("y"), "id")
    )

    result = case.execute()
    expected = pd.DataFrame(
        {
            "id": [1],
            "tz": [pd.Timestamp("2018-01-01 00:00:01", tz="UTC")],
            "no_tz": [pd.Timestamp("2018-01-01 00:00:02")],
            "id_right": [1],
        }
    )
    tm.assert_frame_equal(result, expected)
    snapshot.assert_match(ibis.to_sql(case), "out.sql")


def test_kwargs_passthrough_in_connect():
    con = ibis.connect(
        "postgresql://postgres:postgres@localhost:5432/ibis_testing?sslmode=allow"
    )
    assert con.current_catalog == "ibis_testing"


def test_port():
    # check that we parse and use the port (and then of course fail cuz it's bogus)
    with pytest.raises(PsycoPg2OperationalError):
        ibis.connect("postgresql://postgres:postgres@localhost:1337/ibis_testing")


@h.given(st.integers(min_value=4, max_value=1000))
def test_pgvector_type_load(con, vector_size):
    """
    CREATE TABLE items (id bigserial PRIMARY KEY, embedding vector(3));
    INSERT INTO items (embedding) VALUES ('[1,2,3]'), ('[4,5,6]');
    """
    t = con.table("items")

    assert t.schema() == ibis.schema(
        {
            "id": dt.int64(nullable=False),
            "embedding": dt.unknown,
        }
    )

    result = ["[1,2,3]", "[4,5,6]"]
    assert t.to_pyarrow().column("embedding").to_pylist() == result

    query = f"""
    DROP TABLE IF EXISTS itemsvrandom;
    CREATE TABLE itemsvrandom (id bigserial PRIMARY KEY, embedding vector({vector_size}));
    """

    with con.raw_sql(query):
        pass

    t = con.table("itemsvrandom")

    assert t.schema() == ibis.schema(
        {
            "id": dt.int64(nullable=False),
            "embedding": dt.unknown,
        }
    )

    con.drop_table("itemsvrandom")


def test_name_dtype(con):
    expected_schema = ibis.schema(
        {
            "f_table_catalog": dt.String(nullable=True),
            "f_table_schema": dt.String(nullable=True),
            "f_table_name": dt.String(nullable=True),
            "f_geometry_column": dt.String(nullable=True),
            "coord_dimension": dt.Int32(nullable=True),
            "srid": dt.Int32(nullable=True),
            "type": dt.String(nullable=True),
        }
    )

    assert con.tables.geometry_columns.schema() == expected_schema


def test_infoschema_dtypes(con):
    # information_schema.views

    #   |----------------------------+----------------|
    #   | table_catalog              | sql_identifier |
    #   | table_schema               | sql_identifier |
    #   | table_name                 | sql_identifier |
    #   | view_definition            | character_data |
    #   | check_option               | character_data |
    #   | is_updatable               | yes_or_no      |
    #   | is_insertable_into         | yes_or_no      |
    #   | is_trigger_updatable       | yes_or_no      |
    #   | is_trigger_deletable       | yes_or_no      |
    #   | is_trigger_insertable_into | yes_or_no      |
    #   |----------------------------+----------------|
    #
    views_schema = ibis.schema(
        {
            "table_catalog": dt.String(nullable=True),
            "table_schema": dt.String(nullable=True),
            "table_name": dt.String(nullable=True),
            "view_definition": dt.String(nullable=True),
            "check_option": dt.String(nullable=True),
            "is_updatable": dt.String(nullable=True),
            "is_insertable_into": dt.String(nullable=True),
            "is_trigger_updatable": dt.String(nullable=True),
            "is_trigger_deletable": dt.String(nullable=True),
            "is_trigger_insertable_into": dt.String(nullable=True),
        }
    )

    assert con.table("views", database="information_schema").schema() == views_schema

    # information_schema.sql_sizing

    #   |-----------------+-----------------|
    #   | sizing_id       | cardinal_number |
    #   | sizing_name     | character_data  |
    #   | supported_value | cardinal_number |
    #   | comments        | character_data  |
    #   |-----------------+-----------------|

    sql_sizing_schema = ibis.schema(
        {
            "sizing_id": dt.UInt64(nullable=True),
            "sizing_name": dt.String(nullable=True),
            "supported_value": dt.UInt64(nullable=True),
            "comments": dt.String(nullable=True),
        }
    )

    assert (
        con.table("sql_sizing", database="information_schema").schema()
        == sql_sizing_schema
    )

    # information_schema.triggers has a `created` field with the custom timestamp type
    triggers_created_schema = ibis.schema({"created": dt.Timestamp()})

    assert (
        con.table("triggers", database="information_schema").select("created").schema()
        == triggers_created_schema
    )


def test_password_with_bracket():
    password = f"{IBIS_POSTGRES_PASS}[]"
    quoted_pass = quote_plus(password)
    url = f"postgres://{IBIS_POSTGRES_USER}:{quoted_pass}@{IBIS_POSTGRES_HOST}:{IBIS_POSTGRES_PORT}/{POSTGRES_TEST_DB}"
    with pytest.raises(
        PsycoPg2OperationalError,
        match=f'password authentication failed for user "{IBIS_POSTGRES_USER}"',
    ):
        ibis.connect(url)


def test_create_geospatial_table_with_srid(con):
    name = gen_name("geospatial")
    column_names = string.ascii_lowercase
    column_types = [
        "Point",
        "LineString",
        "Polygon",
        "MultiLineString",
        "MultiPoint",
        "MultiPolygon",
    ]
    schema_string = ", ".join(
        f"{column} geometry({dtype}, 4326)"
        for column, dtype in zip(column_names, column_types)
    )
    con.raw_sql(f"CREATE TEMP TABLE {name} ({schema_string})")
    schema = con.get_schema(name)
    assert schema == ibis.schema(
        {
            column: getattr(dt, dtype)(srid=4326)
            for column, dtype in zip(column_names, column_types)
        }
    )


@pytest.fixture(scope="module")
def enum_table(con):
    name = gen_name("enum_table")
    con.raw_sql("CREATE TYPE mood AS ENUM ('sad', 'ok', 'happy')")
    con.raw_sql(f"CREATE TEMP TABLE {name} (mood mood)")
    yield name
    con.raw_sql(f"DROP TABLE {name}")
    con.raw_sql("DROP TYPE mood")


def test_enum_table(con, enum_table):
    t = con.table(enum_table)
    assert t.mood.type() == dt.unknown


def test_parsing_oid_dtype(con):
    # Load a table that uses the OID type and check that we map it to Int64
    t = con.table("pg_class", database="pg_catalog")
    assert t.oid.type() == ibis.dtype("int64")
